#!/usr/bin/python
# gateway.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (gateway.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: gateway.

This is a replacement for `lib.transport_control` stuff.
Here is place to put `outgoing` files to send to other users in the network.
Also this code receives a `incoming` files from other nodes.

Keeps a list of available `transports` - we use a plug-in system so you can
use external p2p, mech, or centralized networks to transfer files in different ways.

To identify user need to place his ID inside given external network to his public identity file.
So BitDust users can use different ways to communicate and transfer data.
BitDust code will use that module to talk with nodes in the network.

Seems like we faced with such situation -
BitDust software need to work together with other networks software on same machine.
This means we need to communicate between system processes - this is work for plug-ins.
So the idea is to make plug-ins code working inside the main process thread -
they just need to send/receive a short commands to another process in the OS.
So it must be atomic operations (or deferred) and do not use any resources.

And it seems like some plug-ins must have 2 parts:

    * BitDust python module with interface to do peers communications
    * external code to run the given network

Communication between both parts can be done in a different ways, I think XMLRPC can work for most cases.
We need to run XML-RPC on both sides because transport plug-in
must be able to respond to the main process too.

External network can be written in C++ or Java and the second part of the transport plig-in need to
deal with that and be portable.

Different networks provides various functions, something more than just transfer files.
Some of them uses DHT to store data on nodes - we can use that stuff also.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

_PacketLogFileEnabled = False

#------------------------------------------------------------------------------

import os
import time
import optparse

from twisted.web import xmlrpc
from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.defer import Deferred, maybeDeferred, succeed, fail

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.p2p import commands
from bitdust.p2p import p2p_service

from bitdust.lib import nameurl
from bitdust.lib import misc
from bitdust.lib import strng
from bitdust.lib import packetid

from bitdust.system import bpio
from bitdust.system import tmpfile

from bitdust.main import settings
# from bitdust.main import control
from bitdust.main import config
from bitdust.main import events

from bitdust.crypt import signed

from bitdust.contacts import identitycache

from bitdust.transport import callback
from bitdust.transport import packet_in
from bitdust.transport import packet_out

from bitdust.userid import global_id
from bitdust.userid import identity
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_AvailableTransports = {}
_TransportsDict = {}
_LocalListener = None
_XMLRPCListener = None
_XMLRPCPort = None
_XMLRPCURL = ''
_LastTransferID = None
_LastInboxPacketTime = 0
_PacketsTimeOutTask = None
_TransportStateChangedCallbacksList = []
_TransportLogFile = None
_TransportLogFilename = None

#------------------------------------------------------------------------------


def init():
    global _LocalListener
    global _PacketLogFileEnabled
    if _Debug:
        lg.out(4, 'gateway.init')
    if _Debug:
        open_transport_log(settings.TransportLog())
    if _LocalListener:
        lg.warn('local listener already exist')
    else:
        _LocalListener = TransportGateLocalProxy()
    _PacketLogFileEnabled = config.conf().getBool('logs/packet-enabled')


def shutdown():
    """
    Shut down the gateway, need to stop all transports.
    """
    global _LocalListener
    global _XMLRPCListener
    global _XMLRPCPort
    global _XMLRPCURL
    global _PacketLogFileEnabled
    if _Debug:
        lg.out(4, 'gateway.shutdown')
    if _LocalListener:
        _LocalListener = None
    else:
        lg.warn('local listener not exist')
    if _Debug:
        close_transport_log()
    _PacketLogFileEnabled = False


#------------------------------------------------------------------------------


def transport(proto):
    global _TransportsDict
    proto = strng.to_text(proto)
    return _TransportsDict[proto]


def transports():
    global _TransportsDict
    return _TransportsDict


def listener():
    global _LocalListener
    return _LocalListener


def is_ready():
    """
    Return True if gateway is ready to run IN / OUT packets.
    """
    if not listener():
        return False
    return True


def is_installed(proto):
    """
    Return True if given transport is installed.
    """
    return proto in transports()


def can_send(proto):
    return transport(proto).state == 'LISTENING'


def last_inbox_time():
    global _LastInboxPacketTime
    return _LastInboxPacketTime


#------------------------------------------------------------------------------


def start():
    if _Debug:
        lg.out(4, 'gateway.start')
    callback.append_outbox_filter_callback(on_outbox_packet)
    callback.add_finish_file_receiving_callback(on_file_received)
    result = []
    for proto, transp in transports().items():
        if settings.transportIsEnabled(proto):
            if transp.state != 'LISTENING':
                if _Debug:
                    lg.out(4, '    sending "start" to %r' % transp)
                transp.automat('start')
                result.append(proto)
            else:
                if _Debug:
                    lg.out(4, '    %r is ready' % transp)
    reactor.callLater(5, packets_timeout_loop)  # @UndefinedVariable
    return result


def cold_start():
    if _Debug:
        lg.out(4, 'gateway.cold_start : sending "start" only to one transport - most preferable')
    callback.append_outbox_filter_callback(on_outbox_packet)
    callback.add_finish_file_receiving_callback(on_file_received)
    ordered_list = list(transports().keys())
    ordered_list.sort(key=settings.getTransportPriority, reverse=True)
    result = []
    for proto in ordered_list:
        transp = transport(proto)
        if settings.transportIsEnabled(proto):
            if transp.state != 'LISTENING':
                if _Debug:
                    lg.out(4, '    sending "start" to %r only' % transp)
                transp.automat('start')
                result.append(proto)
                break
            else:
                if _Debug:
                    lg.out(4, '    %r is ready, try next one' % transp)
    reactor.callLater(5, packets_timeout_loop)  # @UndefinedVariable
    return result


#------------------------------------------------------------------------------


def stop():
    if _Debug:
        lg.out(4, 'gateway.stop')
    stop_packets_timeout_loop()
    shutdown_all_inbox_packets()
    shutdown_all_outbox_packets()
    result = []
    for proto, transp in transports().items():
        if settings.transportIsEnabled(proto):
            if transp.state != 'OFFLINE':
                if _Debug:
                    lg.out(4, '    send "stop" to %r' % transp)
                transp.automat('stop')
                result.append(proto)
            else:
                if _Debug:
                    lg.out(4, '    %r already stopped' % proto)
    callback.remove_finish_file_receiving_callback(on_file_received)
    callback.remove_outbox_filter_callback(on_outbox_packet)
    return result


#------------------------------------------------------------------------------


def verify():
    ordered_list = list(transports().keys())
    ordered_list.sort(key=settings.getTransportPriority, reverse=True)
    if _Debug:
        lg.out(4, 'gateway.verify sorted list : %r' % ordered_list)
    my_id_obj = my_id.getLocalIdentity()
    resulted = Deferred()
    all_results = {}

    def _verify_transport(proto):
        if _Debug:
            lg.out(_DebugLevel - 2, '    verifying %s_transport' % proto)
        if not settings.transportIsEnabled(proto):
            if _Debug:
                lg.out(_DebugLevel - 2, '    %s_transport is disabled' % proto)
            return succeed(True)
        transp = transport(proto)
        if transp.state == 'OFFLINE':
            if _Debug:
                lg.out(_DebugLevel - 2, '    %s_transport state is OFFLINE' % proto)
            return succeed(True)
        if transp.state != 'LISTENING':
            if _Debug:
                lg.out(_DebugLevel - 2, '    %s_transport state is not LISTENING' % proto)
            return succeed(True)
        transp_result = transp.interface.verify_contacts(my_id_obj)
        if _Debug:
            lg.out(_DebugLevel - 2, '        %s result is %r' % (proto, transp_result))
        if isinstance(transp_result, bool) and transp_result:
            return succeed(True)
        if isinstance(transp_result, bool) and transp_result == False:
            return succeed(False)
        if isinstance(transp_result, Deferred):
            ret = Deferred()
            transp_result.addCallback(lambda result_value: ret.callback(result_value))
            return ret
        lg.warn('incorrect result returned from %s_interface.verify_contacts(): %r' % (proto, transp_result))
        return succeed(False)

    def _on_verified_one(t_result, proto):
        all_results[proto] = t_result
        if _Debug:
            lg.out(_DebugLevel - 2, '        verified %s transport, result=%r' % (proto, t_result))
        if len(all_results) == len(ordered_list):
            resulted.callback((ordered_list, all_results))

    for proto in ordered_list:
        d = _verify_transport(proto)
        d.addCallback(_on_verified_one, proto)

    return resulted


#------------------------------------------------------------------------------


def attach(transport_instance):
    global _TransportsDict
    global _AvailableTransports
    proto = strng.to_text(transport_instance.proto)
    _AvailableTransports[proto] = True
    _TransportsDict[proto] = transport_instance
    if _Debug:
        lg.out(4, 'gateway.attach : %r' % transport_instance)


def detach(transport_instance):
    global _TransportsDict
    global _AvailableTransports
    proto = strng.to_text(transport_instance.proto)
    if proto not in list(_AvailableTransports.keys()):
        lg.warn('transport [%s] not available' % proto)
        return
    if proto not in list(_TransportsDict.keys()):
        lg.warn('transport [%s] not attached' % proto)
        return
    _AvailableTransports.pop(proto)
    _TransportsDict.pop(proto)
    if _Debug:
        lg.out(4, 'gateway.detach : %r' % transport_instance)


#------------------------------------------------------------------------------


def inbox(info):
    """
    1) The protocol modules write to temporary files and gives us that filename
    2) We unserialize
    3) We check that it is for us
    4) We check that it is from one of our contacts.
    5) We use signed.validate() to check signature and that number fields are numbers
    6) Any other sanity checks we can do and if anything funny we toss out the packet
    7) Then change the filename to the PackedID that it should be
      and call the right function(s) for this new packet:
      (encryptedblock, scrubber, remotetester, customerservice, ...)
      to dispatch it to right place(s).
    """
    global _LastInboxPacketTime
    #     if _DoingShutdown:
    #         if _Debug:
    #             lg.out(_DebugLevel, "gateway.inbox ignoring input since _DoingShutdown ")
    #         return None
    if _Debug:
        lg.out(_DebugLevel, 'gateway.inbox [%s]' % info.filename)

    if info.filename == '' or not os.path.exists(info.filename):
        lg.err('bad filename=' + info.filename)
        return None
    try:
        data = bpio.ReadBinaryFile(info.filename)
    except:
        lg.err('gateway.inbox ERROR reading file ' + info.filename)
        return None
    if len(data) == 0:
        lg.err('gateway.inbox ERROR zero byte file from %s://%s' % (info.proto, info.host))
        return None
    if callback.run_finish_file_receiving_callbacks(info, data):
        lg.warn('incoming data of %d bytes was filtered out in file receiving callbacks' % len(data))
        return None
    try:
        newpacket = signed.Unserialize(data)
    except:
        lg.err('gateway.inbox ERROR during Unserialize data from %s://%s' % (info.proto, info.host))
        lg.exc()
        return None
    if newpacket is None:
        lg.warn('newpacket from %s://%s is None' % (info.proto, info.host))
        return None
    # newpacket.Valid() will be called later in the flow in packet_in.handle() method
    try:
        Command = newpacket.Command
        OwnerID = newpacket.OwnerID
        CreatorID = newpacket.CreatorID
        PacketID = newpacket.PacketID
        Date = newpacket.Date
        Payload = newpacket.Payload
        RemoteID = newpacket.RemoteID
        Signature = newpacket.Signature
        packet_sz = len(data)
    except:
        lg.err('gateway.inbox ERROR during Unserialize data from %s://%s' % (info.proto, info.host))
        lg.err('data length=' + str(len(data)))
        lg.exc()
        return None
    _LastInboxPacketTime = time.time()
    if _Debug:
        lg.out(_DebugLevel - 2, 'gateway.inbox [%s] signed by %s|%s (for %s) from %s://%s' % (Command, nameurl.GetName(OwnerID), nameurl.GetName(CreatorID), nameurl.GetName(RemoteID), info.proto, info.host))
    if _PacketLogFileEnabled:
        lg.out(
            0,
            '                \033[1;49;92mINBOX %s(%s) %s %s for %s\033[0m' % (
                newpacket.Command,
                newpacket.PacketID,
                global_id.UrlToGlobalID(newpacket.OwnerID),
                global_id.UrlToGlobalID(newpacket.CreatorID),
                global_id.UrlToGlobalID(newpacket.RemoteID),
            ),
            log_name='packet',
            showtime=True,
        )
    return newpacket


def outbox(
    outpacket,
    wide=False,
    callbacks={},
    target=None,
    route=None,
    response_timeout=None,
    keep_alive=True,
):
    """
    Sends `packet` to the network.

        :param outpacket: an instance of ``signed.Packet``
        :param wide:  set to True if you need to send the packet
                      to all contacts of Remote Identity
        :param callbacks: provide a callback methods to get response
                          here need to provide a callback for given command
                          callback arguments are: (response_packet, info)
        :param target:  if your recipient is not equal to outpacket.RemoteID
        :param route:   dict with parameters, you can manage how to process this packet:
                'packet': <another packet to be send>,
                'proto': <receiver proto>,
                'host': <receiver host>,
                'remoteid': <receiver idurl>,
                'description': <description on the packet>,
        :param response_timeout   None, or integer to indicate how long to wait for an ack

    Returns:
        `None` if data was not sent, no filter was applied
        `Deferred` object if filter was applied but sending was delayed
        `packet_out.PacketOut` object if packet was sent
    """
    if _Debug:
        lg.out(
            _DebugLevel, 'gateway.outbox [%s] signed by %s|%s to %s (%s), wide=%s' % (
                outpacket.Command,
                nameurl.GetName(outpacket.OwnerID),
                nameurl.GetName(outpacket.CreatorID),
                nameurl.GetName(outpacket.RemoteID),
                nameurl.GetName(target),
                wide,
            )
        )
    if _PacketLogFileEnabled:
        lg.out(
            0,
            '\033[1;49;96mOUTBOX %s(%s) %s %s to %s\033[0m' % (
                outpacket.Command,
                outpacket.PacketID,
                global_id.UrlToGlobalID(outpacket.OwnerID),
                global_id.UrlToGlobalID(outpacket.CreatorID),
                global_id.UrlToGlobalID(outpacket.RemoteID),
            ),
            log_name='packet',
            showtime=True,
        )
    return callback.run_outbox_filter_callbacks(
        outpacket,
        wide=wide,
        callbacks=callbacks,
        target=target,
        route=route,
        response_timeout=response_timeout,
        keep_alive=keep_alive,
    )


#------------------------------------------------------------------------------


def make_transfer_ID():
    """
    Generate a unique transfer ID.
    """
    global _LastTransferID
    if _LastTransferID is None:
        _LastTransferID = int(str(int(time.time()*100.0))[4:])
    _LastTransferID += 1
    return _LastTransferID


#------------------------------------------------------------------------------


def connect_to(proto, host):
    if not is_ready():
        return fail(Exception('gateway is not ready'))
    if not is_installed(proto):
        return fail(Exception('transport %r not installed' % proto))
    return transport(proto).call('connect_to', host)


def disconnect_from(proto, host):
    if not is_ready():
        return fail(Exception('gateway is not ready'))
    if not is_installed(proto):
        return fail(Exception('transport %r not installed' % proto))
    return transport(proto).call('disconnect_from', host)


def send_file(remote_idurl, proto, host, filename, description='', pkt_out=None):
    """
    Send a file to remote peer via given transport.

    Args:
        + proto (str): identifier of the transport
        + host (str): remote peer's host comes from identity contact
        + filename (str): local source file to be send
        + description (str): a label for this transfer
        + remote_idurl (idurl): remote user idurl (optional)
    """
    if not is_ready():
        lg.warn('gateway is not ready')
        return False
    if not is_installed(proto):
        lg.warn('transport %r not installed' % proto)
        return False
    filtered = callback.run_file_sending_filter_callbacks(remote_idurl, proto, host, filename, description, pkt_out)
    if filtered is not None:
        return filtered
    result_defer = transport(proto).call('send_file', remote_idurl, filename, host, description)
    callback.run_begin_file_sending_callbacks(result_defer, remote_idurl, proto, host, filename, description, pkt_out)
    return True


def send_file_single(remote_idurl, proto, host, filename, description='', pkt_out=None):
    if not is_ready():
        lg.warn('gateway is not ready')
        return False
    if not is_installed(proto):
        lg.warn('transport %r not installed' % proto)
        return False
    result_defer = transport(proto).call('send_file_single', remote_idurl, filename, host, description)
    callback.run_begin_file_sending_callbacks(result_defer, remote_idurl, proto, host, filename, description, pkt_out)
    return True


def send_keep_alive(proto, host):
    if not is_ready():
        lg.warn('gateway is not ready')
        return False
    if not is_installed(proto):
        lg.warn('transport %r not installed' % proto)
        return False
    transport(proto).call('send_keep_alive', host)
    return True


def list_active_transports():
    if not is_ready():
        return fail(Exception('gateway is not ready'))
    result = []
    for proto, transp in transports().items():
        if settings.transportIsEnabled(proto):
            if transp.state != 'OFFLINE':
                result.append(proto)
    return result


def list_active_sessions(proto):
    if not is_ready():
        return fail(Exception('gateway is not ready'))
    if not is_installed(proto):
        return fail(Exception('transport %r not installed' % proto))
    return transport(proto).call('list_sessions')


def list_active_streams(proto):
    if not is_ready():
        return fail(Exception('gateway is not ready'))
    if not is_installed(proto):
        return fail(Exception('transport %r not installed' % proto))
    return transport(proto).call('list_streams')


def find_active_session(proto, host=None, idurl=None):
    if not is_ready():
        # return fail(Exception('gateway is not ready'))
        lg.warn('gateway is not ready')
        return None
    if not is_installed(proto):
        # return fail(Exception('transport %r not installed' % proto))
        lg.warn('transport %r not installed' % proto)
        return None
    return transport(proto).call('find_session', host, idurl)


def find_active_stream(proto, stream_id=None, transfer_id=None):
    if not is_ready():
        lg.warn('gateway is not ready')
        return None
    if not is_installed(proto):
        lg.warn('transport %r not installed' % proto)
        return None
    return transport(proto).call('find_stream', stream_id=stream_id, transfer_id=transfer_id)


#------------------------------------------------------------------------------


def cancel_input_file(transferID, why=None):
    pkt_in = packet_in.get(transferID)
    assert pkt_in is not None
    if _Debug:
        lg.out(_DebugLevel, 'gateway.cancel_input_file : %s why: %s' % (transferID, why))
    pkt_in.automat('cancel', why)
    return True


def cancel_outbox_file(proto, host, filename, why=None):
    pkt_out, _ = packet_out.search(proto, host, filename)
    if pkt_out is None:
        lg.err('gateway.cancel_outbox_file ERROR packet_out not found: %r' % ((proto, host, filename), ))
        return None
    if _Debug:
        lg.out(_DebugLevel, 'gateway.cancel_outbox_file : %s:%s %s, why: %s' % (proto, host, filename, why))
    pkt_out.automat('cancel', why)


def cancel_outbox_file_by_transfer_id(transferID, why=None):
    pkt_out, _ = packet_out.search_by_transfer_id(transferID)
    if pkt_out is None:
        lg.warn('%s is not found' % str(transferID))
        return False
    if _Debug:
        lg.out(_DebugLevel, 'gateway.cancel_outbox_file_by_transfer_id : %s' % transferID)
    pkt_out.automat('cancel', why)
    return True


#------------------------------------------------------------------------------


def current_bytes_sent():
    res = {}
    for pkt_out in packet_out.queue():
        for item in pkt_out.items:
            if item.transfer_id:
                res[item.transfer_id] = pkt_out.payloadsize
    return res


def current_bytes_received():
    res = {}
    for pkt_in in list(packet_in.inbox_items().values()):
        res[pkt_in.transfer_id] = pkt_in.size
    return res


#------------------------------------------------------------------------------


def shutdown_all_outbox_packets():
    if _Debug:
        lg.out(_DebugLevel, 'gateway.shutdown_all_outbox_packets, %d live objects at the moment' % len(packet_out.queue()))
    for pkt_out in list(packet_out.queue()):
        pkt_out.event('cancel', 'shutdown')


def shutdown_all_inbox_packets():
    if _Debug:
        lg.out(_DebugLevel, 'gateway.shutdown_all_inbox_packets, %d live objects at the moment' % len(list(packet_in.inbox_items().values())))
    for pkt_in in list(packet_in.inbox_items().values()):
        pkt_in.event('cancel', 'shutdown')


#------------------------------------------------------------------------------


def packets_timeout_loop():
    global _PacketsTimeOutTask
    delay = 30
    _PacketsTimeOutTask = reactor.callLater(delay, packets_timeout_loop)  # @UndefinedVariable
    for pkt_in in list(packet_in.inbox_items().values()):
        if pkt_in.is_timed_out():
            if _Debug:
                lg.out(_DebugLevel, 'gateway.packets_timeout_loop %r is timed out: %s' % (pkt_in, pkt_in.timeout))
            pkt_in.automat('cancel', 'timeout')
    for pkt_out in packet_out.queue():
        if pkt_out.is_timed_out():
            if _Debug:
                lg.out(_DebugLevel, 'gateway.packets_timeout_loop %r is timed out: %s' % (pkt_out, pkt_out.timeout))
            pkt_out.automat('cancel', 'timeout')


def stop_packets_timeout_loop():
    global _PacketsTimeOutTask
    if _PacketsTimeOutTask:
        if _PacketsTimeOutTask.active():
            _PacketsTimeOutTask.cancel()
        _PacketsTimeOutTask = None


#------------------------------------------------------------------------------


def monitoring():
    list_pkt_in = []
    for pkt_in in list(packet_in.inbox_items().values()):
        list_pkt_in.append(pkt_in.label)
    list_pkt_out = []
    for pkt_out in packet_out.queue():
        list_pkt_out.append(pkt_out.label)
    if _Debug:
        if transport_log() and list_pkt_out and list_pkt_in:
            dt = time.time() - lg.when_life_begins()
            mn = dt // 60
            sc = dt - mn*60
            transport_log().write(u'%02d:%02d    in: %s   out: %s\n' % (mn, sc, list_pkt_in, list_pkt_out))
            transport_log().flush()


#------------------------------------------------------------------------------


def on_file_received(info, data):
    if _Debug:
        lg.dbg(_DebugLevel, 'received %d bytes in %s' % (len(data), info))
    return False


def on_identity_received(newpacket, send_ack=True):
    """
    A normal node or identity server is sending us a new copy of an identity for a contact of ours.
    Checks that identity is signed correctly.
    This will be also sending requests to cache all sources (other identity servers) holding that identity.
    """
    # TODO:  move to service_gateway
    newxml = newpacket.Payload
    newidentity = identity.identity(xmlsrc=newxml)
    # SECURITY
    # check that identity is signed correctly
    # old public key matches new one
    # this is done in `UpdateAfterChecking()`
    idurl = newidentity.getIDURL()
    if not identitycache.HasKey(idurl):
        lg.info('received new identity %s rev %r' % (idurl, newidentity.getRevisionValue()))
    if my_id.isLocalIdentityReady():
        if newidentity.getPublicKey() == my_id.getLocalIdentity().getPublicKey():
            if newidentity.getRevisionValue() > my_id.getLocalIdentity().getRevisionValue():
                lg.warn('received my own identity from another user, but with higher revision number')
                reactor.callLater(0, my_id.rebuildLocalIdentity, new_revision=newidentity.getRevisionValue() + 1)  # @UndefinedVariable
                return False
    if not identitycache.UpdateAfterChecking(idurl, newxml):
        lg.warn('ERROR has non-Valid identity')
        return False
    latest_identity = id_url.get_latest_ident(newidentity.getPublicKey())
    if latest_identity:
        if latest_identity.getRevisionValue() > newidentity.getRevisionValue():
            # check if received identity is the most recent revision number we ever saw for that remote user
            # in case we saw same identity with higher revision number need to reply with Fail packet and notify user
            # this may happen after identity restore - the user starts counting revision number from 0
            # but other nodes already store previous copies, user just need to jump to the most recent revision number
            lg.warn('received new identity with out-dated revision number %d from %r, known revision is %d' % (newidentity.getRevisionValue(), idurl, latest_identity.getRevisionValue()))
            lg.warn('received identity: %r' % newxml)
            lg.warn('known identity: %r' % latest_identity.serialize())
            ident_packet = signed.Packet(
                Command=commands.Identity(),
                OwnerID=latest_identity.getIDURL(),
                CreatorID=latest_identity.getIDURL(),
                PacketID='identity:%s' % packetid.UniqueID(),
                Payload=latest_identity.serialize(),
                RemoteID=idurl,
            )
            reactor.callLater(0, packet_out.create, outpacket=ident_packet, wide=True, callbacks={}, keep_alive=False)  # @UndefinedVariable
            return False
    # Now that we have ID we can check the packet
    if not newpacket.Valid():
        # If not valid do nothing
        lg.warn('not Valid packet from %s' % idurl)
        return False
    if not send_ack:
        if _Debug:
            lg.dbg(_DebugLevel, '%s  idurl=%s  remoteID=%r  skip sending Ack()' % (newpacket.PacketID, idurl, newpacket.RemoteID))
        return True
    if newpacket.OwnerID == idurl:
        if _Debug:
            lg.dbg(_DebugLevel, '%s  idurl=%s  remoteID=%r  sending wide Ack()' % (newpacket.PacketID, idurl, newpacket.RemoteID))
    else:
        if _Debug:
            lg.dbg(_DebugLevel, '%s  idurl=%s  remoteID=%r  but packet ownerID=%s   sending wide Ack()' % (newpacket.PacketID, idurl, newpacket.RemoteID, newpacket.OwnerID))
    # wide=True : a small trick to respond to all known contacts of the remote user
    reactor.callLater(0, p2p_service.SendAck, newpacket, wide=True)  # @UndefinedVariable
    return True


def on_outbox_packet(outpacket, wide, callbacks, target=None, route=None, response_timeout=None, keep_alive=True):
    started_packets = packet_out.search_similar_packets(outpacket)
    if started_packets:
        for active_packet, _ in started_packets:
            if active_packet.outpacket and active_packet.outpacket.Command in [commands.Ack(), commands.Fail()]:
                continue
            if callbacks:
                for command, cb in callbacks.items():
                    active_packet.set_callback(command, cb)
            lg.warn('skip creating new outbox packet because found similar pending packet: %r' % active_packet)
            return active_packet
    pkt_out = packet_out.create(outpacket, wide, callbacks, target, route, response_timeout, keep_alive)
    # control.request_update([('packet', outpacket.PacketID)])
    return pkt_out


def on_transport_state_changed(transport, oldstate, newstate):
    global _TransportStateChangedCallbacksList
    if _Debug:
        lg.out(_DebugLevel - 2, 'gateway.on_transport_state_changed in %r : %s->%s' % (transport, oldstate, newstate))
    from bitdust.p2p import network_connector
    if network_connector.A():
        network_connector.A('network-transport-state-changed', transport)
    for cb in _TransportStateChangedCallbacksList:
        cb(transport, oldstate, newstate)


def on_transport_initialized(proto, xmlrpcurl=None):
    transport(proto).automat('transport-initialized', xmlrpcurl)
    events.send('gateway-transport-initialized', data=dict(
        proto=proto,
        rpc_url=xmlrpcurl,
    ))
    return True


def on_receiving_started(proto, host, options_modified=None):
    if _Debug:
        lg.out(_DebugLevel - 2, 'gateway.on_receiving_started %r host=%r' % (proto.upper(), host))
    transport(proto).automat('receiving-started', (proto, host, options_modified))
    events.send('gateway-receiving-started', data=dict(
        proto=proto,
        host=host,
        options=options_modified,
    ))
    return True


def on_receiving_failed(proto, error_code=None):
    if _Debug:
        lg.out(_DebugLevel - 2, 'gateway.on_receiving_failed %s    error=[%s]' % (proto.upper(), str(error_code)))
    transport(proto).automat('failed')
    events.send('gateway-receiving-failed', data=dict(
        proto=proto,
        error=error_code,
    ))
    return True


def on_disconnected(proto, result=None):
    if _Debug:
        lg.out(_DebugLevel - 2, 'gateway.on_disconnected %s    result=%s' % (proto.upper(), str(result)))
    if proto in transports():
        transport(proto).automat('stopped')
    events.send('gateway-disconnected', data=dict(
        proto=proto,
        result=result,
    ))
    return True


def on_start_connecting(host):
    return True


def on_session_opened(host, remote_user_id):
    pass


def on_connection_failed(host, error_message=None):
    pass


def on_session_closed(host, remote_user_id, reason=None):
    pass


def on_message_received(host, remote_user_id, data):
    pass


def on_register_file_sending(proto, host, receiver_idurl, filename, size=0, description=''):
    """
    Called from transport plug-in when sending of a single file started towards remote peer.
    Must return a unique transfer ID so plug-in will know that ID.
    After finishing that given transfer - that ID is passed to `unregister_file_sending()`.
    Need to first find existing outgoing packet and register that item.
    """
    if _Debug:
        lg.out(_DebugLevel, 'gateway.on_register_file_sending %s %s to %r' % (filename, description, receiver_idurl))


#     if id_url.field(receiver_idurl).to_bin() == my_id.getIDURL().to_bin():
#         pkt_out, work_item = packet_out.search(proto, host, filename)
#     else:
#         pkt_out, work_item = packet_out.search(proto, host, filename, remote_idurl=receiver_idurl)
    pkt_out, work_item = packet_out.search(proto, host, filename)
    if pkt_out is None:
        lg.warn('skip register file sending, packet_out not found: %r %r %r %r' % (proto, host, os.path.basename(filename), receiver_idurl))
        return None
    transfer_id = make_transfer_ID()
    if _Debug:
        lg.out(_DebugLevel, '... OUT ... %s (%d) send {%s} via [%s] to %s at %s' % (pkt_out.description, transfer_id, os.path.basename(filename), proto, nameurl.GetName(receiver_idurl), host))
    pkt_out.automat('register-item', (proto, host, filename, transfer_id))
    # control.request_update([('stream', transfer_id)])
    return transfer_id


def on_unregister_file_sending(transfer_id, status, bytes_sent, error_message=None):
    """
    Called from transport plug-in after finish sending a single file.
    """
    if transfer_id is None:
        return False
    if _Debug:
        lg.out(_DebugLevel, 'gateway.on_unregister_file_sending %s %s' % (transfer_id, status))
    pkt_out, work_item = packet_out.search_by_transfer_id(transfer_id)
    if pkt_out is None:
        if _Debug:
            lg.out(_DebugLevel, '        %s is not found' % str(transfer_id))
        return False
    pkt_out.automat('unregister-item', (transfer_id, status, bytes_sent, error_message))
    # control.request_update([('stream', transfer_id)])
    if status == 'finished':
        if _Debug:
            lg.out(_DebugLevel, '>>> OUT >>> %s (%d) [%s://%s] %s with %d bytes' % (pkt_out.description, transfer_id, work_item.proto, work_item.host, status.upper(), bytes_sent))
    else:
        if _Debug:
            lg.out(_DebugLevel, '>>> OUT >>> %s (%d) [%s://%s] %s : %s' % (pkt_out.description, transfer_id, work_item.proto, work_item.host, str(status).upper(), error_message))
    return True


def on_cancelled_file_sending(proto, host, filename, size, description='', error_message=None):
    pkt_out, work_item = packet_out.search(proto, host, filename)
    if pkt_out is None:
        if _Debug:
            lg.out(_DebugLevel, 'gateway.on_cancelled_file_sending packet_out %s %s %s not found - IT IS OK' % (proto, host, os.path.basename(filename)))
        return True
    pkt_out.automat('item-cancelled', (
        proto,
        host,
        filename,
        size,
        description,
        error_message,
    ))
    # if pkt_out.outpacket:
    #     control.request_update([('packet', pkt_out.outpacket.PacketID)])
    if _Debug:
        lg.out(_DebugLevel, '>>> OUT >>>  {%s} CANCELLED via [%s] to %s : %s' % (os.path.basename(filename), proto, host, error_message))
    return True


def on_register_file_receiving(proto, host, sender_idurl, filename, size=0):
    """
    Called from transport plug-in when receiving a single file were started
    from some peer.

    Must return a unique transfer ID, create a `FileTransferInfo` object
    and put it into "transfers" list. Plug-in's code must create a
    temporary file and write incoming data into that file.
    """
    transfer_id = make_transfer_ID()
    if _Debug:
        lg.out(_DebugLevel, '... IN ... %d receive {%s} via [%s] from %s at %s' % (transfer_id, os.path.basename(filename), proto, nameurl.GetName(sender_idurl), host))
    incoming_packet = packet_in.create(transfer_id)
    incoming_packet.event('register-item', (proto, host, sender_idurl, filename, size))
    # control.request_update([('stream', transfer_id)])
    return transfer_id


def on_unregister_file_receiving(transfer_id, status, bytes_received, error_message=''):
    """
    Called from transport plug-in after finish receiving a single file.
    """
    pkt_in = packet_in.get(transfer_id)
    if not pkt_in:
        lg.exc(exc_value=Exception('incoming packet with transfer_id=%r not exist' % transfer_id))
        return False
    if _Debug:
        if status == 'finished':
            lg.out(_DebugLevel, '<<< IN <<< (%d) [%s://%s] %s with %d bytes' % (transfer_id, pkt_in.proto, pkt_in.host, status.upper(), bytes_received))
        else:
            lg.out(_DebugLevel, '<<< IN <<< (%d) [%s://%s] %s : %s' % (transfer_id, pkt_in.proto, pkt_in.host, status.upper(), error_message))
    pkt_in.automat('unregister-item', (status, bytes_received, error_message))
    # control.request_update([('stream', transfer_id)])
    return True


#------------------------------------------------------------------------------


def add_transport_state_changed_callback(cb):
    global _TransportStateChangedCallbacksList
    if cb not in _TransportStateChangedCallbacksList:
        _TransportStateChangedCallbacksList.append(cb)


def remove_transport_state_changed_callback(cb):
    global _TransportStateChangedCallbacksList
    if cb in _TransportStateChangedCallbacksList:
        _TransportStateChangedCallbacksList.remove(cb)


#------------------------------------------------------------------------------


def open_transport_log(filename):
    global _TransportLogFile
    global _TransportLogFilename
    if _TransportLogFile:
        return
    _TransportLogFilename = filename
    try:
        _TransportLogFile = open(_TransportLogFilename, 'w')
    except:
        _TransportLogFile = None


def close_transport_log():
    global _TransportLogFile
    if not _TransportLogFile:
        return
    _TransportLogFile.flush()
    _TransportLogFile.close()
    _TransportLogFile = None
    _TransportLogFilename = None


def transport_log():
    global _TransportLogFile
    return _TransportLogFile


#------------------------------------------------------------------------------


class TransportGateLocalProxy():
    """
    A class to handle calls from transport plug-ins in the main thread.
    """
    def __init__(self):
        self.methods = {
            'transport_initialized': on_transport_initialized,
            'receiving_started': on_receiving_started,
            'receiving_failed': on_receiving_failed,
            'disconnected': on_disconnected,
            'start_connecting': on_start_connecting,
            'session_opened': on_session_opened,
            'message_received': on_message_received,
            'connection_failed': on_connection_failed,
            'register_file_sending': on_register_file_sending,
            'unregister_file_sending': on_unregister_file_sending,
            'register_file_receiving': on_register_file_receiving,
            'unregister_file_receiving': on_unregister_file_receiving,
            'cancelled_file_sending': on_cancelled_file_sending,
        }

    def callRemote(self, method, *args):
        m = self.methods.get(method)
        if not m:
            lg.warn('unsupported method: %s' % method)
            return fail(Exception('unsupported method: %s' % method))
        _d = Deferred()

        def _call(meth):
            if _Debug:
                lg.args(_DebugLevel, method=meth, args=args)
            r = maybeDeferred(m, *args)
            r.addCallback(_d.callback)
            r.addErrback(_d.errback)

        reactor.callLater(0, _call, method)  # @UndefinedVariable
        return _d


#------------------------------------------------------------------------------


class TransportGateXMLRPCServer(xmlrpc.XMLRPC):
    """
    XML-RPC server to receive calls from transport plug-ins.
    """
    def __init__(self):
        xmlrpc.XMLRPC.__init__(self, allowNone=True)
        self.methods = {
            'transport_initialized': on_transport_initialized,
            'receiving_started': on_receiving_started,
            'receiving_failed': on_receiving_failed,
            'disconnected': on_disconnected,
            'start_connecting': on_start_connecting,
            'session_opened': on_session_opened,
            'message_received': on_message_received,
            'connection_failed': on_connection_failed,
            'register_file_sending': on_register_file_sending,
            'unregister_file_sending': on_unregister_file_sending,
            'register_file_receiving': on_register_file_receiving,
            'unregister_file_receiving': on_unregister_file_receiving,
            'cancelled_file_sending': on_cancelled_file_sending,
        }

    def lookupProcedure(self, procedurePath):
        try:
            return self.methods[procedurePath]
        except KeyError as e:
            raise xmlrpc.NoSuchFunction(self.NOT_FOUND, 'procedure %s not found: %s' % (procedurePath, e))

    def listProcedures(self):
        return list(self.methods.keys())


#------------------------------------------------------------------------------


def parseCommandLine():
    oparser = optparse.OptionParser()
    oparser.add_option('-d', '--debug', dest='debug', type='int', help='set debug level')
    oparser.set_default('debug', 10)
    oparser.add_option('-t', '--tcpport', dest='tcpport', type='int', help='specify port for TCP transport')
    oparser.set_default('tcpport', settings.getTCPPort())
    oparser.add_option('-u', '--udpport', dest='udpport', type='int', help='specify port for UDP transport')
    oparser.set_default('udpport', settings.getUDPPort())
    oparser.add_option('-p', '--dhtport', dest='dhtport', type='int', help='specify UDP port for DHT network')
    oparser.set_default('dhtport', settings.getDHTPort())
    oparser.add_option('-s', '--packetsize', dest='packetsize', type='int', help='set size of UDP datagrams')
    oparser.set_default('packetsize', 480)
    (options, args) = oparser.parse_args()
    return options, args


#------------------------------------------------------------------------------


def main():
    lg.life_begins()
    bpio.init()
    settings.init()
    misc.init()
    my_id.init()
    identitycache.init()
    from bitdust.crypt import key
    key.InitMyKey()
    (options, args) = parseCommandLine()
    settings.override('transport.transport-tcp.transport-tcp-port', options.tcpport)
    settings.override('transport.transport-udp.transport-udp-port', options.udpport)
    settings.override('network.network-dht-port', options.dhtport)
    lg.set_debug_level(options.debug)
    tmpfile.init()
    if True:
        import bitdust.lib.udp
        bitdust.lib.udp.listen(options.udpport)
        import bitdust.dht.dht_service
        bitdust.dht.dht_service.init(options.dhtport)
    reactor.addSystemEventTrigger('before', 'shutdown', shutdown)  # @UndefinedVariable
    init()
    start()
    globals()['num_in'] = 0

    def _in(a, b, c, d):
        lg.out(2, 'INBOX %d : %r' % (globals()['num_in'], a))
        globals()['num_in'] += 1
        return False

    callback.insert_inbox_callback(0, _in)
    if len(args) > 0:
        globals()['num_out'] = 0

        def _s():
            p = signed.Packet(commands.Data(), my_id.getIDURL(), my_id.getIDURL(), my_id.getIDURL(), bpio.ReadBinaryFile(args[1]), args[0])
            outbox(p, wide=True)
            lg.out(2, 'OUTBOX %d : %r' % (globals()['num_out'], p))
            globals()['num_out'] += 1

        old_state_changed = transport('udp').state_changed

        def new_state_changed(oldstate, newstate, event, *args, **kwargs):
            old_state_changed(oldstate, newstate, event, *args, **kwargs)
            if newstate == 'LISTENING':
                reactor.callLater(1, _s)  # @UndefinedVariable

        transport('udp').state_changed = new_state_changed
        # t = task.LoopingCall(_s)
        # reactor.callLater(5, t.start, 60, True)
        # reactor.callLater(2, t.stop)

    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# msgtypes.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
import six

import hashlib
import random

from . import encoding  # @UnresolvedImport


class Message(object):
    """ Base class for messages - all "unknown" messages use this class """
    def __init__(self, rpcID, nodeID):
        self.id = encoding.to_text(rpcID)
        self.nodeID = encoding.to_text(nodeID)

    def __repr__(self, *args, **kwargs):
        return str(self)


class RequestMessage(Message):
    """
    Message containing an RPC request.
    """
    def __init__(self, nodeID, method, methodArgs, rpcID=None, layerID=0):
        if rpcID is None:
            hsh = hashlib.sha1()
            hsh.update(str(random.getrandbits(255)).encode())
            rpcID = hsh.hexdigest()
        Message.__init__(self, rpcID, nodeID)
        self.request = method
        self.args = methodArgs
        self.layerID = layerID

    def __str__(self):
        return '<RequestMessage(%d) %s %s with %r>' % (self.layerID, self.id[:6], self.nodeID[:6], self.args)


class QuestionMessage(Message):
    """
    Message containing an RPC request from a node which do not want to be included in DHT layer.
    """
    def __init__(self, nodeID, method, methodArgs, rpcID=None, layerID=0):
        if rpcID is None:
            hsh = hashlib.sha1()
            hsh.update(str(random.getrandbits(255)).encode())
            rpcID = hsh.hexdigest()
        Message.__init__(self, rpcID, nodeID)
        self.request = method
        self.args = methodArgs
        self.layerID = layerID

    def __str__(self):
        return '<QuestionMessage(%d) %s %s with %r>' % (self.layerID, self.id[:6], self.nodeID[:6], self.args)


class ResponseMessage(Message):
    """
    Message containing the result from a successful RPC request.
    """
    def __init__(self, rpcID, nodeID, response, layerID=0):
        Message.__init__(self, rpcID, nodeID)
        self.response = response
        self.layerID = layerID

    def __str__(self):
        return '<ResponseMessage(%d) %s %s with %r>' % (self.layerID, self.id[:6], self.nodeID[:6], self.response)


class ErrorMessage(ResponseMessage):
    """
    Message containing the error from an unsuccessful RPC request.
    """
    def __init__(self, rpcID, nodeID, exceptionType, errorMessage, layerID=0):
        ResponseMessage.__init__(self, rpcID, nodeID, errorMessage, layerID=layerID)
        if isinstance(exceptionType, type):
            self.exceptionType = '%s.%s' % (exceptionType.__module__, exceptionType.__name__)
        else:
            self.exceptionType = exceptionType
            if isinstance(self.exceptionType, six.binary_type):
                self.exceptionType = self.exceptionType.decode()

    def __str__(self):
        return '<ErrorMessage(%d) %s %s with %r>' % (self.layerID, self.id[:6], self.nodeID[:6], self.exceptionType)



#!/usr/bin/python
# git_proc.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (git_proc.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: git_proc.

A code for all platforms to perform source code updates from official Git repo at:

   http://dev.bitdust.io/code/public.git
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from io import open

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import sys
import time

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in git_proc.py')

from twisted.internet import protocol

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import deploy

from bitdust.main import settings
from bitdust.main import events

#------------------------------------------------------------------------------

_CurrentProcess = None
_FirstRunDelay = 1200
_LoopInterval = 3600*6
_ShedulerTask = None

#------------------------------------------------------------------------------


def write2log(txt):
    out_file = open(settings.UpdateLogFilename(), 'a')
    out_file.write(strng.to_text(txt))
    out_file.close()


#------------------------------------------------------------------------------


def init():
    lg.out(4, 'git_proc.init')
    if os.environ.get('BITDUST_GIT_SYNC_SKIP', '0') == '1':
        return
    reactor.callLater(0, loop, first_start=True)  # @UndefinedVariable


def shutdown():
    lg.out(4, 'git_proc.shutdown')
    global _ShedulerTask
    if _ShedulerTask is not None:
        if _ShedulerTask.active():
            _ShedulerTask.cancel()
            lg.out(4, '    loop stopped')
        _ShedulerTask = None


#------------------------------------------------------------------------------


def sync_callback(result):
    lg.out(6, 'git_proc.sync_callback: %s' % result)

    if result == 'code-fetched':
        events.send('source-code-fetched', data=dict())
    elif result == 'up-to-date':
        events.send('source-code-up-to-date', data=dict())
    else:
        events.send('source-code-update-error', data=dict(result=result))

    try:
        from bitdust.system import tray_icon
        if result == 'error':
            # tray_icon.draw_icon('error')
            # reactor.callLater(5, tray_icon.restore_icon)
            return
        elif result == 'code-fetched':
            tray_icon.set_icon('updated')
            return
    except:
        pass


def run_sync():
    lg.out(6, 'git_proc.run_sync')
    reactor.callLater(0, sync, sync_callback, update_method='reset')  # @UndefinedVariable
    reactor.callLater(0, loop)  # @UndefinedVariable


def loop(first_start=False):
    global _ShedulerTask
    lg.out(4, 'git_proc.loop')
    if first_start:
        nexttime = time.time() + _FirstRunDelay
    else:
        nexttime = time.time() + _LoopInterval
    # DEBUG
    # nexttime = time.time() + 10.0
    delay = nexttime - time.time()
    if delay < 0:
        lg.warn('delay=%s %s %s' % (str(delay), nexttime, time.time()))
        delay = 0
    lg.out(6, 'git_proc.loop run_sync will start after %s minutes' % str(delay/60.0))
    _ShedulerTask = reactor.callLater(delay, run_sync)  # @UndefinedVariable


#------------------------------------------------------------------------------


def sync(callback_func=None, update_method='rebase'):
    """
    Runs commands and process stdout and stderr to recogneze the result:

        `git fetch --all -v`
        `git rebase origin/master -v`  or  `git reset --hard origin/master`

    """
    src_dir_path = bpio.getExecutableDir()
    expected_src_dir = os.path.join(deploy.default_base_dir_portable(), 'src')
    if bpio.portablePath(src_dir_path) != bpio.portablePath(expected_src_dir):
        if _Debug:
            lg.out(_DebugLevel, 'git_proc.sync SKIP, non standard sources location: %r' % src_dir_path)
        return

    def _reset_done(response, error, retcode, result):
        if callback_func is None:
            return
        callback_func(result)

    def _rebase_done(response, error, retcode, result):
        if callback_func is None:
            return
        if retcode != 0:
            result = 'sync-error'
        else:
            if response.count(b'Changes from') or response.count(b'Fast-forwarded'):
                result = 'code-fetched'
            else:
                result = 'up-to-date'
        callback_func(result)

    def _fetch_done(response, error, retcode):
        if retcode != 0:
            if callback_func:
                callback_func('sync-error')
            return
        result = 'sync-error'
        if response.count(b'Unpacking') or \
            (response.count(b'master') and response.count(b'->')) or \
            response.count(b'Updating') or \
            response.count(b'Receiving') or \
                response.count(b'Counting'):
            result = 'new-code'
        if update_method == 'reset':
            run([
                'reset',
                '--hard',
                'origin/master',
            ], callback=lambda resp, err, ret: _reset_done(resp, err, ret, result))
        elif update_method == 'rebase':
            run(['rebase', 'origin/master', '-v'], callback=lambda resp, err, ret: _rebase_done(resp, err, ret, result))
        else:
            raise Exception('invalid update method: %s' % update_method)

    run(['fetch', '--all', '-v'], callback=_fetch_done)


#------------------------------------------------------------------------------


def run(cmdargs, base_dir=None, git_bin=None, env=None, callback=None):
    if _Debug:
        lg.out(_DebugLevel, 'git_proc.run')
    base_dir = base_dir or bpio.getExecutableDir()
    if bpio.Windows():
        cmd = [
            'git',
        ] + cmdargs
        if git_bin:
            git_exe = git_bin
        else:
            git_exe = bpio.portablePath(os.path.join(base_dir, '..', 'git', 'bin', 'git.exe'))
        if not os.path.isfile(git_exe):
            if _Debug:
                lg.out(_DebugLevel, '    not found git.exe, try to run from shell')
            try:
                response, error, retcode = execute_in_shell(cmd, base_dir=base_dir)
            except:
                response = ''
                error = ''
                retcode = 1
            if callback:
                callback(response, error, retcode)
            return
        if _Debug:
            lg.out(_DebugLevel, '    found git in %s' % git_exe)
        cmd = [
            git_exe,
        ] + cmdargs
    else:
        cmd = [
            git_bin or 'git',
        ] + cmdargs
    execute(cmd, callback=callback, base_dir=base_dir, env=env)


#------------------------------------------------------------------------------


def execute_in_shell(cmdargs, base_dir=None):
    global _CurrentProcess
    from bitdust.system import nonblocking
    import subprocess
    if _Debug:
        lg.out(_DebugLevel, 'git_proc.execute_in_shell: "%s"' % (' '.join(cmdargs)))
    write2log('EXECUTE in shell: %s, base_dir=%s' % (cmdargs, base_dir))
    _CurrentProcess = nonblocking.Popen(
        cmdargs,
        shell=True,
        cwd=bpio.portablePath(base_dir),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    result = _CurrentProcess.communicate()
    out_data = result[0]
    err_data = result[1]
    write2log('STDOUT:\n%s\nSTDERR:\n%s\n' % (out_data, err_data))
    returncode = _CurrentProcess.returncode
    if _Debug:
        lg.out(_DebugLevel, 'git_proc.execute_in_shell returned: %s, stdout bytes: %d, stderr bytes: %d' % (returncode, len(out_data), len(err_data)))
    return (out_data, err_data, returncode)  # _CurrentProcess


#------------------------------------------------------------------------------


class GitProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, callbacks=[]):
        self.callbacks = callbacks
        self.out = b''
        self.err = b''

    def errReceived(self, inp):
        self.err += inp
        for line in inp.splitlines():
            if _Debug:
                lg.out(_DebugLevel, '[git:err]: %s' % strng.to_text(line))

    def outReceived(self, inp):
        self.out += inp
        for line in inp.splitlines():
            if _Debug:
                lg.out(_DebugLevel, '[git:out]: %s' % strng.to_text(line))

    def processEnded(self, reason):
        if _Debug:
            lg.out(_DebugLevel, 'git process FINISHED : %s' % reason.value.exitCode)
        for cb in self.callbacks:
            cb(self.out, self.err, reason.value.exitCode)


def execute(cmdargs, base_dir=None, process_protocol=None, env=None, callback=None):
    global _CurrentProcess
    if _Debug:
        lg.out(_DebugLevel, 'git_proc.execute: "%s" in %s' % (' '.join(cmdargs), base_dir))
    write2log('EXECUTE: %s, base_dir=%s' % (cmdargs, base_dir))
    executable = cmdargs[0]
    if bpio.Windows():
        from twisted.internet import _dumbwin32proc
        real_CreateProcess = _dumbwin32proc.win32process.CreateProcess  # @UndefinedVariable

        def fake_createprocess(_appName, _commandLine, _processAttributes, _threadAttributes, _bInheritHandles, creationFlags, _newEnvironment, _currentDirectory, startupinfo):
            import win32con  # @UnresolvedImport
            import subprocess
            flags = win32con.CREATE_NO_WINDOW
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            startupinfo.wShowWindow = subprocess.SW_HIDE
            return real_CreateProcess(_appName, _commandLine, _processAttributes, _threadAttributes, _bInheritHandles, flags, _newEnvironment, _currentDirectory, startupinfo)

        setattr(_dumbwin32proc.win32process, 'CreateProcess', fake_createprocess)

    if process_protocol is None:
        process_protocol = GitProcessProtocol(callbacks=[
            lambda out, err, ret_code: write2log('STDOUT:\n%s\nSTDERR:\n%s\n' % (out, err)),
            callback,
        ])
    try:
        _CurrentProcess = reactor.spawnProcess(  # @UndefinedVariable
            process_protocol, executable, cmdargs, path=base_dir, env=env
        )
    except:
        lg.exc()
        return None
    if bpio.Windows():
        setattr(_dumbwin32proc.win32process, 'CreateProcess', real_CreateProcess)
    return _CurrentProcess


#------------------------------------------------------------------------------

if __name__ == '__main__':
    bpio.init()
    lg.set_debug_level(18)

    def _result(res):
        print('RESULT:', res)
        reactor.stop()  # @UndefinedVariable

    reactor.callWhenRunning(sync, _result)  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable



#!/usr/bin/env python
# storage_contract.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (storage_contract.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
from bitdust.coins.contract_chain_consumer import _Debug
"""
.. module:: storage_contract.

"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime
from bitdust.lib import jsn

from bitdust.system import local_fs
from bitdust.system import bpio

from bitdust.main import config
from bitdust.main import settings
from bitdust.main import events

from bitdust.storage import accounting

from bitdust.blockchain import bismuth_wallet

from bitdust.userid import id_url

#------------------------------------------------------------------------------


def get_customer_contracts_dir(customer_idurl):
    customer_contracts_prefix = id_url.field(customer_idurl).unique_name()
    return os.path.join(settings.ServiceDir('service_supplier_contracts'), customer_contracts_prefix)


def get_current_customer_contract(customer_idurl):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, path=customer_contracts_dir)
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    current_customer_contract_path = os.path.join(customer_contracts_dir, 'current')
    if not os.path.isfile(current_customer_contract_path):
        return None
    json_data = jsn.loads_text(local_fs.ReadTextFile(current_customer_contract_path))
    if not json_data:
        return None
    if not accounting.verify_storage_contract(json_data):
        lg.err('current storage contract with %r is not valid' % customer_idurl)
        return None
    if 'my_position' in json_data:
        # TODO: remove later...
        json_data['ecc_position'] = json_data.pop('my_position')
    return json_data


def get_customer_transaction(customer_idurl, tx_id):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, tx_id=tx_id)
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    current_customer_contract_path = os.path.join(customer_contracts_dir, 'current')
    if not os.path.isfile(current_customer_contract_path):
        return None
    json_data = jsn.loads_text(local_fs.ReadTextFile(current_customer_contract_path))
    if not json_data:
        return None
    if not accounting.verify_storage_contract(json_data):
        lg.err('current storage contract with %r is not valid' % customer_idurl)
        return None
    if 'my_position' in json_data:
        # TODO: remove later...
        json_data['ecc_position'] = json_data.pop('my_position')
    return json_data


def is_current_customer_contract_active(customer_idurl):
    current_contract = get_current_customer_contract(customer_idurl)
    if not current_contract:
        return False
    now = utime.utcnow_to_sec1970()
    if now > utime.unpack_time(current_contract['complete_after']):
        return False
    return True


def list_customer_contracts(customer_idurl):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    customer_contracts = {}
    current_contract = None
    latest_contract = None
    latest_contract_state = None
    latest_contract_started_time = -1
    latest_completed_contract = None
    latest_completed_contract_time = -1
    latest_paid_contract = None
    latest_paid_contract_time = -1
    completed_contracts_total_GBH_value = 0
    paid_contracts_total_GBH_value = 0
    counter = 0
    for fname in os.listdir(customer_contracts_dir):
        if fname == 'current':
            current_contract = jsn.loads_text(local_fs.ReadTextFile(os.path.join(customer_contracts_dir, 'current')))
            if not accounting.verify_storage_contract(current_contract):
                current_contract = None
                lg.err('current storage contract is invalid')
            if 'my_position' in current_contract:
                # TODO: remove later...
                current_contract['ecc_position'] = current_contract.pop('my_position')
            continue
        try:
            started_time = int(fname.split('.')[0])
            contract_state = fname.split('.')[1]
        except:
            lg.exc()
            continue
        contract_path = os.path.join(customer_contracts_dir, fname)
        json_data = jsn.loads_text(local_fs.ReadTextFile(contract_path))
        if not accounting.verify_storage_contract(json_data):
            lg.err('invalid storage contract found: %r' % fname)
            continue
        if 'my_position' in json_data:
            # TODO: remove later...
            json_data['ecc_position'] = json_data.pop('my_position')
        json_data['filename'] = fname
        json_data['state'] = contract_state
        customer_contracts[started_time] = json_data
        counter += 1
        if started_time > latest_contract_started_time:
            latest_contract_started_time = started_time
            latest_contract = json_data
            latest_contract_state = contract_state
        if contract_state == 'completed':
            completed_contracts_total_GBH_value += json_data['value']
            if started_time > latest_completed_contract_time:
                latest_completed_contract = json_data
        elif contract_state == 'cancelled':
            completed_contracts_total_GBH_value += json_data['completed_value']
        elif contract_state == 'paid':
            paid_contracts_total_GBH_value += json_data['value']
            if started_time > latest_paid_contract_time:
                latest_paid_contract = json_data
        else:
            raise Exception('unknown state of the contract')
    customer_contracts['current'] = current_contract
    customer_contracts['latest'] = latest_contract
    customer_contracts['latest_state'] = latest_contract_state
    customer_contracts['latest_paid_contract'] = latest_paid_contract
    customer_contracts['latest_completed_contract'] = latest_completed_contract
    customer_contracts['completed_value'] = completed_contracts_total_GBH_value
    customer_contracts['paid_value'] = paid_contracts_total_GBH_value
    customer_contracts['count'] = counter
    if _Debug:
        lg.args(_DebugLevel, completed_value=completed_contracts_total_GBH_value, paid_value=paid_contracts_total_GBH_value)
    return customer_contracts


def prepare_customer_contract(customer_idurl, details):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, path=customer_contracts_dir)
    now = utime.utcnow_to_sec1970()
    started_time = now
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    current_customer_contract_path = os.path.join(customer_contracts_dir, 'current')
    if os.path.isfile(current_customer_contract_path):
        current_contract = jsn.loads_text(local_fs.ReadTextFile(current_customer_contract_path))
        if current_contract:
            if not accounting.verify_storage_contract(current_contract):
                current_contract = None
                lg.err('current storage contract with %r is not valid' % customer_idurl)
            else:
                if now < utime.unpack_time(current_contract['complete_after']):
                    # SCENARIO 2: there is already a contract started with this customer and not yet completed
                    if _Debug:
                        lg.dbg(_DebugLevel, 'valid customer contract with %r already exists' % customer_idurl)
                    return change_current_customer_contract(customer_idurl, details)
                # SCENARIO 3,4,6: found that previous contract was active but finished now
                lg.warn('current storage contract with %r already ended' % customer_idurl)
                complete_current_customer_contract(customer_idurl)
    customer_contracts_list_and_details = list_customer_contracts(customer_idurl)
    latest_contract = customer_contracts_list_and_details['latest']
    latest_contract_state = customer_contracts_list_and_details['latest_state']
    latest_paid_contract = customer_contracts_list_and_details['latest_paid_contract']
    latest_completed_contract = customer_contracts_list_and_details['latest_completed_contract']
    completed_value = customer_contracts_list_and_details['completed_value']
    billing_period_seconds = settings.SupplierContractBillingPeriodDays()*24*60*60
    initial_duration_hours = config.conf().getInt('services/supplier-contracts/initial-duration-hours')
    new_raise_factor = config.conf().getFloat('services/supplier-contracts/duration-raise-factor')
    new_duration_hours = None
    new_sequence_number = 0
    if not latest_contract:
        # SCENARIO 1: this is a new customer - there were no contracts signed yet
        started_time = now
        new_duration_hours = initial_duration_hours
        new_pay_before_time = started_time + int(billing_period_seconds/2.0)
        new_sequence_number = 1
    else:
        latest_sequence_number = latest_contract.get('sequence_number', -1)
        if latest_sequence_number < 0:
            latest_sequence_number = customer_contracts_list_and_details['count']
        new_sequence_number = latest_sequence_number + 1
        if latest_contract_state == 'cancelled':
            cancelled_contract_filename = latest_contract.pop('filename', None)
            if now < utime.unpack_time(latest_contract['complete_after']):
                # SCENARIO 15,16,17: latest contract was cancelled, but customer wants to restart it
                try:
                    os.remove(os.path.join(customer_contracts_dir, cancelled_contract_filename))
                except:
                    lg.exc()
                return start_current_customer_contract(
                    customer_idurl=customer_idurl,
                    details=details,
                    started_time=utime.unpack_time(latest_contract['started']),
                    complete_after_time=utime.unpack_time(latest_contract['complete_after']),
                    pay_before_time=utime.unpack_time(latest_contract['pay_before']),
                    duration_hours=latest_contract['duration_hours'],
                    raise_factor=latest_contract['raise_factor'],
                    sequence_number=new_sequence_number,
                )
            if latest_completed_contract:
                if now - initial_duration_hours*60*60 > utime.unpack_time(latest_completed_contract['pay_before']):
                    # SCENARIO 18, 19: latest contract was cancelled and already expired, customer wants to start a new contract
                    # other contract was completed, but was not paid
                    lg.warn('customer %r yet did not pay for previously completed contract' % customer_idurl)
                    return {
                        'deny': True,
                        'reason': 'unpaid',
                        'value': completed_value,
                    }
                # SCENARIO 11,13: latest contract was cancelled and already expired, other contracts was completed already
                try:
                    os.remove(os.path.join(customer_contracts_dir, cancelled_contract_filename))
                except:
                    lg.exc()
                # rename cancelled contract file to "<started_time>.completed"
                contract_path_new = os.path.join(customer_contracts_dir, '{}.completed'.format(utime.unpack_time(latest_contract['started'])))
                local_fs.WriteTextFile(contract_path_new, jsn.dumps(latest_contract))
                started_time = now
                new_duration_hours = initial_duration_hours
                new_pay_before_time = utime.unpack_time(latest_completed_contract['pay_before'])
            else:
                if latest_paid_contract:
                    # SCENARIO 12: latest contract was cancelled and already expired, other contracts was paid
                    try:
                        os.remove(os.path.join(customer_contracts_dir, cancelled_contract_filename))
                    except:
                        lg.exc()
                    # rename cancelled contract file to "<started_time>.completed"
                    contract_path_new = os.path.join(customer_contracts_dir, '{}.completed'.format(utime.unpack_time(latest_contract['started'])))
                    local_fs.WriteTextFile(contract_path_new, jsn.dumps(latest_contract))
                    started_time = now
                    new_duration_hours = initial_duration_hours
                    new_pay_before_time = started_time + int(billing_period_seconds/2.0)
                else:
                    # SCENARIO 10: latest contract was cancelled and already expired, no other contracts found
                    # rename cancelled contract file to "<started_time>.completed"
                    contract_path_new = os.path.join(customer_contracts_dir, '{}.completed'.format(utime.unpack_time(latest_contract['started'])))
                    local_fs.WriteTextFile(contract_path_new, jsn.dumps(latest_contract))
                    started_time = now
                    new_duration_hours = initial_duration_hours
                    new_pay_before_time = started_time + int(billing_period_seconds/2.0)
        else:
            started_time = utime.unpack_time(latest_contract['complete_after'])
            new_duration_hours = int(latest_contract['duration_hours']*latest_contract['raise_factor'])
            if latest_contract_state == 'completed':
                if latest_paid_contract:
                    new_pay_before_time = utime.unpack_time(latest_paid_contract['started']) + latest_paid_contract['duration_hours']*60*60 + billing_period_seconds
                    if new_pay_before_time < started_time + new_duration_hours*60*60:
                        # SCENARIO 9: the previous contract is completed and some contracts already paid, but there is still no trust to this customer
                        lg.warn('customer %r paid before, but yet did not pay for previously completed contracts' % customer_idurl)
                        return {
                            'deny': True,
                            'reason': 'unpaid',
                            'value': completed_value,
                        }
                    else:
                        # SCENARIO 7: no active contract, the previos contract is completed and there is also a paid contract before
                        pass
                else:
                    new_pay_before_time = utime.unpack_time(latest_contract['pay_before'])
                    if new_pay_before_time < now + new_duration_hours*60*60:
                        # SCENARIO 8: the previous contract is completed, but there is no trust to this customer
                        lg.warn('customer %r yet did not pay for previously completed contract' % customer_idurl)
                        return {
                            'deny': True,
                            'reason': 'unpaid',
                            'value': completed_value,
                        }
                    else:
                        # SCENARIO 5: no active contract, the previos contract is completed but there were no payments yet done
                        pass
            elif latest_contract_state == 'paid':
                # SCENARIO 14: currently there is no active contract, but the previos contract is completed and paid
                new_duration_hours = initial_duration_hours
                new_pay_before_time = started_time + billing_period_seconds
            else:
                raise Exception('unexpected contract state: %r' % latest_contract_state)
    return start_current_customer_contract(
        customer_idurl=customer_idurl,
        details=details,
        started_time=started_time,
        complete_after_time=started_time + new_duration_hours*60*60,
        pay_before_time=new_pay_before_time,
        duration_hours=new_duration_hours,
        raise_factor=new_raise_factor,
        sequence_number=new_sequence_number,
    )


def start_current_customer_contract(customer_idurl, details, started_time, complete_after_time, pay_before_time, duration_hours, raise_factor, sequence_number):
    json_data = {
        'started': utime.pack_time(started_time),
        'complete_after': utime.pack_time(complete_after_time),
        'pay_before': utime.pack_time(pay_before_time),
        'value': float(duration_hours)*(details['allocated_bytes']/(1024.0*1024.0*1024.0)),
        'allocated_bytes': details['allocated_bytes'],
        'duration_hours': duration_hours,
        'ecc_position': details['ecc_position'],
        'ecc_map': details['ecc_map'],
        'raise_factor': raise_factor,
        'sequence_number': sequence_number,
        'wallet_address': bismuth_wallet.my_wallet_address(),
    }
    local_fs.WriteTextFile(os.path.join(get_customer_contracts_dir(customer_idurl), 'current'), jsn.dumps(json_data))
    json_data['customer'] = customer_idurl
    if _Debug:
        lg.args(_DebugLevel, c=json_data)
    events.send('storage-contract-started', data=dict(contract=json_data))
    return json_data


def cancel_customer_contract(customer_idurl):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, path=customer_contracts_dir)
    if not os.path.isdir(customer_contracts_dir):
        return False
    current_customer_contract_path = os.path.join(customer_contracts_dir, 'current')
    if not os.path.isfile(current_customer_contract_path):
        return False
    current_contract = jsn.loads_text(local_fs.ReadTextFile(current_customer_contract_path))
    if not current_contract:
        return False
    if not accounting.verify_storage_contract(current_contract):
        lg.err('current contract with %r is not valid' % customer_idurl)
        return False
    cancelled_time = utime.utcnow_to_sec1970()
    completed_duration_hours = (cancelled_time - utime.unpack_time(current_contract['started']))/(60.0*60.0)
    completed_value = float(completed_duration_hours)*(current_contract['allocated_bytes']/(1024.0*1024.0*1024.0))
    current_contract['completed_value'] = completed_value
    current_contract['cancelled'] = utime.pack_time(cancelled_time)
    # rename "current" file to "<started_time>.cancelled"
    try:
        os.remove(current_customer_contract_path)
    except:
        lg.exc()
    contract_path_new = os.path.join(customer_contracts_dir, '{}.cancelled'.format(utime.unpack_time(current_contract['started'])))
    local_fs.WriteTextFile(contract_path_new, jsn.dumps(current_contract))
    if _Debug:
        lg.args(_DebugLevel, old_path=current_customer_contract_path, new_path=contract_path_new)
    current_contract['customer'] = customer_idurl
    events.send('storage-contract-cancelled', data=dict(contract=current_contract))
    return True


def change_current_customer_contract(customer_idurl, details):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, path=customer_contracts_dir)
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    current_customer_contract_path = os.path.join(customer_contracts_dir, 'current')
    if not os.path.isfile(current_customer_contract_path):
        lg.err('current storage contract with %r not found' % customer_idurl)
        return {
            'deny': True,
            'reason': 'current storage contract not found',
        }
    current_contract = jsn.loads_text(local_fs.ReadTextFile(current_customer_contract_path))
    if not current_contract:
        lg.err('current storage contract with %r read failed' % customer_idurl)
        return {
            'deny': True,
            'reason': 'current storage contract not found',
        }
    if details.get('ecc_map'):
        current_contract['ecc_map'] = details['ecc_map']
    if details.get('ecc_position') is not None:
        current_contract['ecc_position'] = details['ecc_position']
    if details['allocated_bytes'] != current_contract['allocated_bytes']:
        current_value = current_contract['value']
        new_duration_hours = int(current_value/(details['allocated_bytes']/(1024.0*1024.0*1024.0)))
        new_complete_after_time = utime.unpack_time(current_contract['started']) + new_duration_hours*60*60
        if new_complete_after_time > utime.unpack_time(current_contract['pay_before']):
            return {
                'deny': True,
                'reason': 'contract duration change limit exceeded',
            }
        current_contract['allocated_bytes'] = details['allocated_bytes']
        current_contract['duration_hours'] = new_duration_hours
        current_contract['complete_after'] = utime.pack_time(new_complete_after_time)
    if _Debug:
        lg.args(_DebugLevel, c=current_contract)
    local_fs.WriteTextFile(current_customer_contract_path, jsn.dumps(current_contract))
    current_contract['customer'] = customer_idurl
    events.send('storage-contract-changed', data=dict(contract=current_contract))
    return current_contract


def complete_current_customer_contract(customer_idurl):
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, path=customer_contracts_dir)
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    current_customer_contract_path = os.path.join(customer_contracts_dir, 'current')
    if not os.path.isfile(current_customer_contract_path):
        return False
    current_contract = jsn.loads_text(local_fs.ReadTextFile(current_customer_contract_path))
    if not current_contract:
        return False
    if not accounting.verify_storage_contract(current_contract):
        lg.err('current contract with %r is not valid' % customer_idurl)
        return False
    # rename "current" file to "<started_time>.completed"
    contract_path_new = os.path.join(customer_contracts_dir, '{}.completed'.format(utime.unpack_time(current_contract['started'])))
    if _Debug:
        lg.args(_DebugLevel, old_path=current_customer_contract_path, new_path=contract_path_new)
    os.rename(current_customer_contract_path, contract_path_new)
    current_contract['customer'] = customer_idurl
    events.send('storage-contract-completed', data=dict(contract=current_contract))
    return True


def verify_all_current_customers_contracts():
    rejected_customers = []
    now = utime.utcnow_to_sec1970()
    if not os.path.isdir(settings.ServiceDir('service_supplier_contracts')):
        return []
    for customer_unique_name in os.listdir(settings.ServiceDir('service_supplier_contracts')):
        try:
            customer_idurl = id_url.field(id_url.unique_names(customer_unique_name)[0])
        except:
            lg.exc()
            continue
        contracts_list = list_customer_contracts(customer_idurl)
        latest_contract = contracts_list['latest']
        if contracts_list['current']:
            if now > utime.unpack_time(contracts_list['current']['complete_after']):
                lg.warn('current storage contract with %r already ended' % customer_idurl)
                complete_current_customer_contract(customer_idurl)
                latest_contract = contracts_list['current']
        if not latest_contract:
            rejected_customers.append(customer_idurl)
            lg.warn('rejecting customer %r because of missing contract' % customer_idurl)
        else:
            if latest_contract.get('cancelled'):
                lg.warn('rejecting customer %r because of cancelled contract' % customer_idurl)
                rejected_customers.append(customer_idurl)
            else:
                next_complete_after_time = utime.unpack_time(latest_contract['complete_after']) + latest_contract['raise_factor']*latest_contract['duration_hours']*60*60
                if now > next_complete_after_time:
                    lg.warn('rejecting customer %r because of finished contract' % customer_idurl)
                    rejected_customers.append(customer_idurl)
    return rejected_customers


def verify_accept_storage_payment(tx):
    try:
        customer_prefix, sequence_numbers = tx['openfield'].split(' ')
        int(tx['block_height'])
    except:
        lg.exc()
        return 0
    try:
        sequence_numbers = list(map(int, sequence_numbers.split(',')))
    except:
        return 0
    recently_completed_contracts = []
    try:
        customer_idurl = id_url.field(id_url.unique_names(customer_prefix)[0])
    except:
        lg.exc()
        return 0
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, unique_name=customer_idurl.unique_name())
    if customer_idurl.unique_name() != customer_prefix:
        lg.err('customer %r identity name do not match' % customer_idurl)
        return 0
    contracts_list = list_customer_contracts(customer_idurl)
    for contract_started_time in contracts_list.keys():
        if isinstance(contract_started_time, int):
            json_data = contracts_list[contract_started_time]
            if json_data['state'] == 'completed':
                if int(json_data['sequence_number']) in sequence_numbers:
                    recently_completed_contracts.append(json_data)
    if _Debug:
        lg.args(_DebugLevel, customer_prefix=customer_prefix, sequence_numbers=sequence_numbers, recently_completed_contracts=len(recently_completed_contracts))
    if not recently_completed_contracts:
        return 0
    customer_contracts_dir = get_customer_contracts_dir(customer_idurl)
    count = 0
    for json_data in recently_completed_contracts:
        # rename "<started_time>.completed" file to "<started_time>.paid"
        completed_contract_path = os.path.join(customer_contracts_dir, json_data['filename'])
        paid_contract_path = os.path.join(customer_contracts_dir, '{}.paid'.format(utime.unpack_time(json_data['started'])))
        if _Debug:
            lg.args(_DebugLevel, block_height=int(tx['block_height']), old_path=completed_contract_path, new_path=paid_contract_path)
        os.rename(completed_contract_path, paid_contract_path)
        paid_contract = jsn.loads_text(local_fs.ReadTextFile(paid_contract_path))
        paid_contract['customer'] = customer_idurl
        events.send('storage-contract-paid', data=dict(contract=paid_contract))
        count += 1
    return count


def scan_recent_storage_transactions():
    transactions_dir_path = os.path.join(settings.ServiceDir('bismuth_blockchain'), 'transactions')
    if not os.path.isdir(transactions_dir_path):
        return
    for block_height, operation, _, filename in bismuth_wallet.my_known_transactions():
        if operation == 'storage':
            filepath = os.path.join(transactions_dir_path, filename)
            try:
                tx = jsn.loads_text(local_fs.ReadTextFile(filepath))
                int(tx['block_height'])
            except:
                lg.exc()
                continue
            accepted = verify_accept_storage_payment(tx)
            if _Debug:
                lg.args(_DebugLevel, block_height=block_height, accepted=accepted)



#!/usr/bin/python
# hashes.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (hashes.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: hashes.


"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10
_CryptoLog = None

#------------------------------------------------------------------------------

try:
    from Cryptodome.Hash import MD5
    from Cryptodome.Hash import SHA1
    from Cryptodome.Hash import SHA256
except:
    from Crypto.Hash import MD5  # @UnresolvedImport @Reimport
    from Crypto.Hash import SHA1  # @UnresolvedImport @Reimport
    from Crypto.Hash import SHA256  # @UnresolvedImport @Reimport

#------------------------------------------------------------------------------

from bitdust.lib import strng
from bitdust.logs import lg

#------------------------------------------------------------------------------


def md5(inp, hexdigest=False, return_object=False):
    global _CryptoLog
    # if _CryptoLog is None:
    #     _CryptoLog = os.environ.get('CRYPTO_LOG') == '1'
    if not strng.is_bin(inp):
        raise ValueError('input must by byte string')
    h = MD5.new(inp)
    if _Debug:
        if _CryptoLog:
            lg.args(_DebugLevel, hexdigest=h.hexdigest())
    if return_object:
        return h
    if hexdigest:
        return strng.to_bin(h.hexdigest())
    return h.digest()


def sha1(inp, hexdigest=False, return_object=False):
    global _CryptoLog
    # if _CryptoLog is None:
    #     _CryptoLog = os.environ.get('CRYPTO_LOG') == '1'
    if not strng.is_bin(inp):
        raise ValueError('input must by byte string')
    h = SHA1.new(inp)
    if _Debug:
        if _CryptoLog:
            lg.args(_DebugLevel, hexdigest=h.hexdigest())
    if return_object:
        return h
    if hexdigest:
        return strng.to_bin(h.hexdigest())
    return h.digest()


def sha256(inp, hexdigest=False, return_object=False):
    global _CryptoLog
    # if _CryptoLog is None:
    #     _CryptoLog = os.environ.get('CRYPTO_LOG') == '1'
    if not strng.is_bin(inp):
        raise ValueError('input must by byte string')
    h = SHA256.new(inp)
    if _Debug:
        if _CryptoLog:
            lg.args(_DebugLevel, hexdigest=h.hexdigest())
    if return_object:
        return h
    if hexdigest:
        return strng.to_bin(h.hexdigest())
    return h.digest()



#!/usr/bin/env python
# data_receiver.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (data_receiver.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: data_receiver
.. role:: red

BitDust data_receiver() Automat

EVENTS:
    * :red:`init`
    * :red:`input-stream-closed`
    * :red:`input-stream-opened`
    * :red:`shutdown`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

from bitdust.automats import automat

from bitdust.transport import callback

#------------------------------------------------------------------------------

_DataReceiver = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _DataReceiver
    if event is None and not args:
        return _DataReceiver
    if _DataReceiver is None:
        _DataReceiver = DataReceiver(
            name='data_receiver',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
            publish_events=False,
        )
    if event is not None:
        _DataReceiver.automat(event, *args, **kwargs)
    return _DataReceiver


#------------------------------------------------------------------------------


class DataReceiver(automat.Automat):
    """
    This class implements all the functionality of ``data_receiver()`` state machine.
    """
    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READY'
                self.doInit(*args, **kwargs)
                self.StreamsCounter = 0
        #---READY---
        elif self.state == 'READY':
            if event == 'input-stream-opened':
                self.state = 'RECEIVING'
                self.StreamsCounter += 1
            elif event == 'shutdown':
                self.state = 'CLOSE'
                self.doDestroyMe(*args, **kwargs)
        #---RECEIVING---
        elif self.state == 'RECEIVING':
            if event == 'input-stream-closed' and self.StreamsCounter > 1:
                self.StreamsCounter -= 1
            elif event == 'input-stream-closed' and self.StreamsCounter == 1:
                self.state = 'READY'
                self.StreamsCounter = 0
            elif event == 'input-stream-opened':
                self.StreamsCounter += 1
            elif event == 'shutdown':
                self.state = 'CLOSE'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSE---
        elif self.state == 'CLOSE':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        callback.add_begin_file_receiving_callback(self._on_begin_file_receiving)
        callback.add_finish_file_receiving_callback(self._on_finish_file_receiving)

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        callback.remove_finish_file_receiving_callback(self._on_finish_file_receiving)
        callback.remove_begin_file_receiving_callback(self._on_begin_file_receiving)
        self.destroy()
        global _DataReceiver
        del _DataReceiver
        _DataReceiver = None

    def _on_begin_file_receiving(self, pkt_in):
        self.event('input-stream-opened', pkt_in)

    def _on_finish_file_receiving(self, pkt_in, data):
        self.event('input-stream-closed', (
            pkt_in,
            data,
        ))



#!/usr/bin/python
# miner.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (miner.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: miner.

Data to be store in global "coins" storage:

For example this should be written for a coin mined by supplier to declare storage sharing:

    + signer: idurl of this node
    + partner: idurl of given customer
    + type: storage (can be also : cpu, traffic, hosting, etc.)
    + start: time in UTC
    + end: time in UTC
    + amount: in megabytes
    + price: 1.0 by default


Customer in opposite will mine such "coins":

    + signer: idurl of this node
    + partner: idurl of supplier
    + type: storage
    + start: time in UTC
    + end: time in UTC
    + amount: in megabytes
    + quality: 1.0 if supplier passed all validation tests for this period


So we have a contract between customer and supplier and both sides declare periodically how it is going
by "mining" a safe "crypto-coins" and put some details into the coin.


We have also "status coins" or "ping coins". When a node get connected to other node it first sends a Identity
packet to "ping" remote side and waits for Ack packet in response. A correct Ack means that remote side is
online at the moment. So we want to declare that info to be able to keep track of how long this node
was online in the past and so how reliable he was.
So the first node will mine a "ping coin" and put such info into it:

    + signer: idurl of this node
    + partner: idurl of remote node
    + type: ack
    + start: time in UTC when he sent Identity packet to remote peer
    + end: time in UTC when he receives Ack packet from remote peer

And remote node should declare that as well to confirm this, this is sort of "pong coin":

    + signer: idurl of this node
    + partner: idurl of the first node who send a "ping" packet
    + type: identity
    + start: time in UTC when this node receives a packet
    + end: <empty>
"""

from __future__ import absolute_import

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

import datetime

#------------------------------------------------------------------------------

_CoinsMinerNode = None
_MyStartedContracts = []
_MyFinishedContracts = []

#------------------------------------------------------------------------------


def node():
    global _CoinsMinerNode
    return _CoinsMinerNode


def started_contracts():
    global _MyStartedContracts
    return _MyStartedContracts


def finished_contracts():
    global _MyFinishedContracts
    return _MyFinishedContracts


#------------------------------------------------------------------------------


def init():
    global _CoinsMinerNode
    if _CoinsMinerNode:
        return
    _CoinsMinerNode = CoinsMinerNode()


def shutdown():
    global _CoinsMinerNode
    if _CoinsMinerNode is None:
        return
    del _CoinsMinerNode
    _CoinsMinerNode = None


#------------------------------------------------------------------------------


def inbox_packet(newpacket, info, status, error_message):
    if status != 'finished':
        return False
    if not node():
        return False
    return node().inbox_packet(newpacket, info)


#------------------------------------------------------------------------------


def start_contract(typ, partner):
    contract = Contract(
        type=typ,
        partner=partner,
        start=datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'),
        end=None,
    )
    started_contracts(contract)


def finish_contract(typ, partner, **kwargs):
    found = None
    for contract in started_contracts():
        if contract.type == typ and contract.partner == partner:
            found = contract
            break
    if not found:
        return False
    started_contracts().remove(found)
    found.end = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
    for key, value in kwargs.items():
        setattr(found, key, value)
    finished_contracts().append(found)


#------------------------------------------------------------------------------


class Contract(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


#------------------------------------------------------------------------------


class CoinsMinerNode(object):
    def inbox_packet(self, newpacket, info):
        return False

    def mine_and_send(self, data):
        from bitdust.transport import gateway
        outpacket = ''
        gateway.outbox(outpacket)



#!/usr/bin/env python
# customer_space.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (customer_space.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: customer_space.

"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import time
import base64

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import packetid
from bitdust.lib import serialization
from bitdust.lib import jsn

from bitdust.system import bpio

from bitdust.main import settings
from bitdust.main import events

from bitdust.contacts import contactsdb

from bitdust.services import driver

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

from bitdust.storage import accounting

from bitdust.crypt import signed
from bitdust.crypt import my_keys

from bitdust.transport import gateway
from bitdust.transport import callback

from bitdust.stream import p2p_queue

from bitdust.dht import dht_service
from bitdust.dht import dht_records
from bitdust.dht import known_nodes

from bitdust.supplier import list_files
from bitdust.supplier import local_tester

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_SupplierFileModifiedLatest = {}
_SupplierFileModifiedNotifyTasks = {}

#------------------------------------------------------------------------------


def init():
    callback.append_inbox_callback(on_inbox_packet_received)
    events.add_subscriber(on_identity_url_changed, 'identity-url-changed')
    events.add_subscriber(on_customer_accepted, 'existing-customer-accepted')
    events.add_subscriber(on_customer_accepted, 'new-customer-accepted')
    events.add_subscriber(on_customer_terminated, 'existing-customer-denied')
    events.add_subscriber(on_customer_terminated, 'existing-customer-terminated')
    space_dict, _ = accounting.read_customers_quotas()
    for customer_idurl in contactsdb.customers():
        known_customer_meta_info = contactsdb.get_customer_meta_info(customer_idurl)
        # yapf: disable
        events.send('existing-customer-accepted', data=dict(
            idurl=customer_idurl,
            allocated_bytes=space_dict.get(customer_idurl.to_bin()),
            ecc_map=known_customer_meta_info.get('ecc_map'),
            position=known_customer_meta_info.get('position'),
        ))
        # yapf: enable
    if driver.is_on('service_entangled_dht'):
        connect_suppliers_dht_layer()
    else:
        lg.warn('service service_entangled_dht is OFF')
    events.add_subscriber(on_dht_layer_connected, 'dht-layer-connected')


def shutdown():
    events.remove_subscriber(on_dht_layer_connected, 'dht-layer-connected')
    if driver.is_on('service_entangled_dht'):
        dht_service.suspend(layer_id=dht_records.LAYER_SUPPLIERS)
    events.remove_subscriber(on_customer_accepted, 'existing-customer-accepted')
    events.remove_subscriber(on_customer_accepted, 'new-customer-accepted')
    events.remove_subscriber(on_customer_terminated, 'existing-customer-denied')
    events.remove_subscriber(on_customer_terminated, 'existing-customer-terminated')
    events.remove_subscriber(on_identity_url_changed, 'identity-url-changed')
    callback.remove_inbox_callback(on_inbox_packet_received)


#------------------------------------------------------------------------------


def connect_suppliers_dht_layer():
    known_seeds = known_nodes.nodes()
    d = dht_service.open_layer(
        layer_id=dht_records.LAYER_SUPPLIERS,
        seed_nodes=known_seeds,
        connect_now=True,
        attach=True,
    )
    d.addCallback(on_suppliers_dht_layer_connected)
    d.addErrback(lambda *args: lg.err(str(args)))
    return d


#------------------------------------------------------------------------------


def register_customer_key(customer_public_key_id, customer_public_key):
    """
    Check/refresh/store customer public key locally.
    """
    if not customer_public_key_id or not customer_public_key:
        lg.warn('customer public key was not provided in the request')
        return False
    customer_public_key_id = my_keys.latest_key_id(customer_public_key_id)
    if my_keys.is_key_registered(customer_public_key_id):
        known_customer_public_key = my_keys.get_public_key_raw(customer_public_key_id)
        if known_customer_public_key == customer_public_key:
            lg.info('customer public key %r already known and public key is matching' % customer_public_key_id)
        else:
            lg.warn('rewriting customer public key %r' % customer_public_key_id)
            my_keys.erase_key(customer_public_key_id)
    key_id, key_object = my_keys.read_key_info(customer_public_key)
    if not my_keys.register_key(key_id, key_object):
        lg.err('failed to register customer public key: %r' % customer_public_key_id)
        return False
    lg.info('new customer public key registered: %r' % customer_public_key_id)
    return True


#------------------------------------------------------------------------------


def verify_ownership(newpacket, raise_exception=False):
    """
    At that point packet creator is already verified via signature,
    but creator could be not authorized to store data on that node.
    Based on owner ID, creator ID and packet ID decision must be made what to do with the packet.
    Returns customer IDURL and authorized IDURL of the user who should receive the Ack() or (None, None, ) if not authorized.
    """
    # SECURITY
    owner_idurl = newpacket.OwnerID
    creator_idurl = newpacket.CreatorID
    owner_id = owner_idurl.to_id()
    creator_id = creator_idurl.to_id()
    packet_key_alias, packet_owner_id, _ = packetid.SplitKeyOwnerData(newpacket.PacketID)
    customer_idurl = global_id.glob2idurl(packet_owner_id)
    packet_key_id = my_keys.latest_key_id(my_keys.make_key_id(packet_key_alias, creator_idurl, creator_glob_id=packet_owner_id))
    packet_key_id_registered = my_keys.is_key_registered(packet_key_id)
    if _Debug:
        lg.args(_DebugLevel, owner_id=owner_id, creator_id=creator_id, customer_id=packet_owner_id, key_registered=packet_key_id_registered)
    if newpacket.Command == commands.Data():
        if id_url.is_the_same(owner_idurl, creator_idurl):
            if contactsdb.is_customer(creator_idurl):
                if _Debug:
                    lg.dbg(_DebugLevel, 'OK, scenario 1: customer is sending own data to own supplier')
                return customer_idurl, creator_idurl
            if packet_key_id_registered:
                if _Debug:
                    lg.dbg(_DebugLevel, 'OK, scenario 10: data sender is not my customer but packet key is registered')
                return customer_idurl, creator_idurl
            lg.err('FAIL, scenario 6: data sender is not my customer and also packet key is not registered')
            if raise_exception:
                raise Exception('non-authorized user is trying to store data')
            return None, None
        if contactsdb.is_customer(creator_idurl):
            if _Debug:
                lg.dbg(_DebugLevel, 'OK, scenario 2: customer wants to store data for someone else on own supplier')
            # TODO: check that, why do we need that?
            return customer_idurl, creator_idurl
        if packet_owner_id == owner_id:
            if contactsdb.is_customer(owner_idurl):
                if packet_key_id_registered:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'OK, scenario 3: another authorized user is sending data to customer to be stored on the supplier')
                    return customer_idurl, creator_idurl
        if _Debug:
            lg.dbg(_DebugLevel, 'non-authorized user is trying to store data on the supplier')
        return None, None
    if newpacket.Command in [commands.DeleteFile(), commands.DeleteBackup()]:
        if owner_idurl == creator_idurl:
            if contactsdb.is_customer(creator_idurl):
                if _Debug:
                    lg.dbg(_DebugLevel, 'OK, scenario 4: customer wants to remove already stored data on own supplier')
                return customer_idurl, creator_idurl
            lg.err('FAIL, scenario 7: non-authorized user is trying to erase data owned by customer from the supplier')
            if raise_exception:
                raise Exception('non-authorized user is trying to erase data owned by customer from the supplier')
            return None, None
        if contactsdb.is_customer(creator_idurl):
            # TODO: check that, why do we need that?
            if _Debug:
                lg.dbg(_DebugLevel, 'OK, scenario 8: customer wants to erase existing data that belongs to someone else but stored on the supplier')
            return customer_idurl, creator_idurl
        if packet_owner_id == owner_id:
            if contactsdb.is_customer(owner_idurl):
                if packet_key_id_registered:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'OK, scenario 5: another authorized user wants to remove already stored data from the supplier')
                    return customer_idurl, creator_idurl
        lg.err('non-authorized user is trying to erase data on the supplier')
        return None, None
    if driver.is_enabled('service_proxy_server'):
        if _Debug:
            lg.dbg(_DebugLevel, 'IGNORE, scenario 9: received Data() was not authorized, but proxy router service is enabled')
        return None, None
    # TODO:
    # scenario 11: make possible to set "active" flag True/False for any key
    # this way customer can make virtual location available for other user but in read-only mode
    raise Exception('scenario not implemented yet, received %r' % newpacket)


#------------------------------------------------------------------------------


def make_filename(customerGlobID, filePath, keyAlias=None):
    keyAlias = keyAlias or 'master'
    customerDirName = str(customerGlobID)
    customersDir = settings.getCustomersFilesDir()
    if not os.path.exists(customersDir):
        if _Debug:
            lg.dbg(_DebugLevel, 'making a new folder: %s' % customersDir)
        bpio._dir_make(customersDir)
    ownerDir = os.path.join(customersDir, customerDirName)
    if not os.path.exists(ownerDir):
        if _Debug:
            lg.dbg(_DebugLevel, 'making a new folder: %s' % ownerDir)
        bpio._dir_make(ownerDir)
    keyAliasDir = os.path.join(ownerDir, keyAlias)
    if not os.path.exists(keyAliasDir):
        if _Debug:
            lg.dbg(_DebugLevel, 'making a new folder: %s' % keyAliasDir)
        bpio._dir_make(keyAliasDir)
    filename = os.path.join(keyAliasDir, filePath)
    return filename


def make_valid_filename(customerIDURL, glob_path):
    """
    Must be a customer, and then we make full path filename for where this
    packet is stored locally.
    """
    keyAlias = glob_path['key_alias'] or 'master'
    filePath = glob_path['path']
    customerGlobID = glob_path['customer']
    if not customerGlobID:
        lg.warn('customer id is empty: %r' % glob_path)
        return ''
    if filePath != settings.BackupIndexFileName():
        if not packetid.Valid(filePath):  # SECURITY
            lg.warn('invalid file path')
            return ''
    # if not contactsdb.is_customer(customerIDURL):  # SECURITY
    #     lg.warn("%s is not my customer" % (customerIDURL))
    # if customerGlobID:
    #     if glob_path['idurl'] != customerIDURL:
    #         lg.warn('making filename for another customer: %s != %s' % (
    #             glob_path['idurl'], customerIDURL))
    filename = make_filename(customerGlobID, filePath, keyAlias)
    return filename


#------------------------------------------------------------------------------


def do_notify_supplier_file_modified(key_alias, remote_path, action, customer_idurl, authorized_idurl):
    global _SupplierFileModifiedNotifyTasks
    global _SupplierFileModifiedLatest
    task_id = '{}_{}${}'.format(action, key_alias, customer_idurl.to_id())
    latest_event_time = _SupplierFileModifiedLatest.get(task_id)
    current_task = _SupplierFileModifiedNotifyTasks.get(task_id)
    if not latest_event_time or (time.time() - latest_event_time > 60):
        if _Debug:
            lg.args(_DebugLevel, t=task_id, cur=current_task)
        if current_task:
            if not current_task.called and not current_task.cancelled:
                current_task.cancel()
            _SupplierFileModifiedNotifyTasks.pop(task_id)
        _SupplierFileModifiedLatest[task_id] = time.time()
        events.send('supplier-file-modified', data=dict(
            action=action,
            remote_path=remote_path,
            key_alias=key_alias,
            authorized_idurl=authorized_idurl,
            customer_idurl=customer_idurl,
            supplier_idurl=my_id.getIDURL(),
        ))
        return
    new_delay = latest_event_time + 60 + 1 - time.time()
    if _Debug:
        lg.args(_DebugLevel, t=task_id, delay=new_delay, cur=current_task)
    if current_task:
        if not current_task.called and not current_task.cancelled:
            current_task.cancel()
        _SupplierFileModifiedNotifyTasks.pop(task_id)
    _SupplierFileModifiedNotifyTasks[task_id] = reactor.callLater(  # @UndefinedVariable
        new_delay,
        do_notify_supplier_file_modified,
        key_alias,
        remote_path,
        action,
        customer_idurl,
        authorized_idurl,
    )


#------------------------------------------------------------------------------


def on_data(newpacket):
    if id_url.is_the_same(newpacket.OwnerID, my_id.getIDURL()):
        # this Data belong to us, SKIP
        return False


#     if not contactsdb.is_customer(newpacket.OwnerID):
#         # SECURITY
#         # TODO: process files from another customer : glob_path['idurl']
#         lg.warn("skip, %s not a customer, packetID=%s" % (newpacket.OwnerID, newpacket.PacketID))
#         # p2p_service.SendFail(newpacket, 'not a customer')
#         return False
    glob_path = global_id.ParseGlobalID(newpacket.PacketID)
    if not glob_path['path']:
        # backward compatible check
        glob_path = global_id.ParseGlobalID(my_id.getGlobalID('master') + ':' + newpacket.PacketID)
    if not glob_path['path']:
        lg.err('got incorrect PacketID')
        # p2p_service.SendFail(newpacket, 'incorrect path')
        return False
    remote_path = glob_path['path']
    key_alias = glob_path['key_alias']
    customer_idurl, authorized_idurl = verify_ownership(newpacket)
    if authorized_idurl is None or customer_idurl is None:
        if _Debug:
            lg.dbg(_DebugLevel, 'ownership verification failed for %r' % newpacket)
        # p2p_service.SendFail(newpacket, 'ownership verification failed')
        return False
    filename = make_valid_filename(newpacket.OwnerID, glob_path)
    if not filename:
        lg.warn('got empty filename, bad customer or wrong packetID?')
        # p2p_service.SendFail(newpacket, 'empty filename')
        return False
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        try:
            bpio._dirs_make(dirname)
        except:
            lg.err('can not create sub dir %s' % dirname)
            p2p_service.SendFail(newpacket, 'write error', remote_idurl=authorized_idurl)
            return False
    new_data = newpacket.Serialize()
    donated_bytes = settings.getDonatedBytes()
    accounting.check_create_customers_quotas(donated_bytes)
    space_dict, _ = accounting.read_customers_quotas()
    known_customers_quoats = list(space_dict.keys())
    if id_url.is_not_in(customer_idurl, known_customers_quoats, as_field=False, as_bin=True):
        lg.err('customer space is broken, no info about donated space can be found for %s' % newpacket)
        p2p_service.SendFail(newpacket, 'customer space is broken, no info found about donated space', remote_idurl=authorized_idurl)
        return False
    used_space_dict = accounting.read_customers_usage()
    known_customers_usage = list(used_space_dict.keys())
    if id_url.is_in(customer_idurl, known_customers_usage, as_field=False, as_bin=True):
        try:
            bytes_used_by_customer = int(used_space_dict[customer_idurl.to_bin()])
            bytes_donated_to_customer = int(space_dict[customer_idurl.to_bin()])
            if bytes_donated_to_customer - bytes_used_by_customer < len(new_data):
                lg.warn('no free space left for customer data for %s' % customer_idurl)
                p2p_service.SendFail(newpacket, 'no free space left for customer data', remote_idurl=authorized_idurl)
                return False
        except:
            lg.exc()
    data_exists = not os.path.exists(filename)
    data_changed = True
    if not data_exists:
        if remote_path == settings.BackupIndexFileName():
            current_data = bpio.ReadBinaryFile(filename)
            if current_data == new_data:
                lg.warn('skip rewriting existing file %s' % filename)
                data_changed = False
    if data_changed:
        if not bpio.WriteBinaryFile(filename, new_data):
            lg.err('can not write to %s' % str(filename))
            p2p_service.SendFail(newpacket, 'write error', remote_idurl=authorized_idurl)
            return False
    # Here Data() packet was stored as it is on supplier node (current machine)
    del new_data
    sz = len(newpacket.Payload)
    p2p_service.SendAck(newpacket, response=strng.to_text(sz), remote_idurl=authorized_idurl)
    reactor.callLater(0, local_tester.TestSpaceTime)  # @UndefinedVariable
    if key_alias != 'master' and data_changed:
        if remote_path == settings.BackupIndexFileName():
            do_notify_supplier_file_modified(key_alias, remote_path, 'write', customer_idurl, authorized_idurl)
        else:
            if packetid.BlockNumber(newpacket.PacketID) == 0:
                do_notify_supplier_file_modified(key_alias, remote_path, 'write', customer_idurl, authorized_idurl)
    if _Debug:
        lg.args(_DebugLevel, sz=sz, fn=filename, remote_idurl=authorized_idurl, pid=newpacket.PacketID)
    return True


def on_retrieve(newpacket):
    # external customer must be able to request
    # TODO: add validation of public key
    # if not contactsdb.is_customer(newpacket.OwnerID):
    #     lg.err("had unknown customer %s" % newpacket.OwnerID)
    #     p2p_service.SendFail(newpacket, 'not a customer')
    #     return False
    glob_path = global_id.ParseGlobalID(newpacket.PacketID)
    if not glob_path['path']:
        # backward compatible check
        glob_path = global_id.ParseGlobalID(my_id.getGlobalID('master') + ':' + newpacket.PacketID)
    if not glob_path['path']:
        lg.err('got incorrect PacketID')
        p2p_service.SendFail(newpacket, 'incorrect path', remote_idurl=newpacket.CreatorID)
        return False
    if not glob_path['idurl']:
        lg.warn('no customer global id found in PacketID: %s' % newpacket.PacketID)
        p2p_service.SendFail(newpacket, 'incorrect retrieve request', remote_idurl=newpacket.CreatorID)
        return False
    key_id = glob_path['key_id']
    recipient_idurl = newpacket.OwnerID
    if newpacket.CreatorID != glob_path['idurl'] and newpacket.CreatorID != newpacket.OwnerID:
        # SECURITY
        lg.warn('one of customers requesting a Data from another customer!')
        if not my_keys.is_key_registered(key_id):
            lg.warn('key %s is not registered' % key_id)
            p2p_service.SendFail(newpacket, 'key not registered', remote_idurl=newpacket.CreatorID)
            return False
        verified = False
        if _Debug:
            lg.args(_DebugLevel, Payload=newpacket.Payload)
        try:
            json_payload = serialization.BytesToDict(newpacket.Payload, keys_to_text=True, values_to_text=True)
            test_sample_bin = base64.b64decode(json_payload['t'])
            test_signature_bin = strng.to_bin(json_payload['s'])
            verified = my_keys.verify(key_id, test_sample_bin, test_signature_bin)
        except:
            lg.exc()
            return False
        if not verified:
            lg.warn('request is not authorized, test sample signature verification failed')
            return False
        # requester signed the test sample with the private key and we verified the signature with the public key
        # now we checked the signature of the test sample and can be sure that requester really possess the same key
        recipient_idurl = newpacket.CreatorID
    filename = make_valid_filename(newpacket.OwnerID, glob_path)
    if not filename:
        filename = make_valid_filename(glob_path['idurl'], glob_path)
        # if True:
        # TODO: settings.getCustomersDataSharingEnabled() and
        # SECURITY
        # TODO: add more validations for receiver idurl
        # recipient_idurl = glob_path['idurl']
        # filename = make_valid_filename(glob_path['idurl'], glob_path)
    if not filename:
        lg.warn('had empty filename')
        p2p_service.SendFail(newpacket, 'empty filename', remote_idurl=recipient_idurl)
        return False
    if not os.path.exists(filename):
        lg.warn('did not found requested file locally : %s' % filename)
        p2p_service.SendFail(newpacket, 'did not found requested file locally', remote_idurl=recipient_idurl)
        return False
    if not os.access(filename, os.R_OK):
        lg.warn('no read access to requested packet %s' % filename)
        p2p_service.SendFail(newpacket, 'failed reading requested file', remote_idurl=recipient_idurl)
        return False
    data = bpio.ReadBinaryFile(filename)
    if not data:
        lg.warn('empty data on disk %s' % filename)
        p2p_service.SendFail(newpacket, 'empty data on disk', remote_idurl=recipient_idurl)
        return False
    stored_packet = signed.Unserialize(data)
    sz = len(data)
    del data
    if stored_packet is None:
        lg.warn('Unserialize failed, not Valid packet %s' % filename)
        p2p_service.SendFail(newpacket, 'unserialize failed', remote_idurl=recipient_idurl)
        return False
    if not stored_packet.Valid():
        lg.warn('Stored packet is not Valid %s' % filename)
        p2p_service.SendFail(newpacket, 'stored packet is not valid', remote_idurl=recipient_idurl)
        return False
    if stored_packet.Command != commands.Data():
        lg.warn('sending back packet which is not a Data')
    # here Data() packet is sent back as it is...
    # that means outpacket.RemoteID=my_id.getIDURL() - it was addressed to that node and stored as it is
    # need to take that into account: every time you receive Data() packet
    # it can be not a new Data(), but the old data returning back as a response to Retreive() packet
    # to solve the issue we will create a new Data() packet
    # which will be addressed directly to recipient and "wrap" stored data inside it
    payload = stored_packet.Serialize()
    routed_packet = signed.Packet(
        Command=commands.Data(),
        OwnerID=stored_packet.OwnerID,
        CreatorID=my_id.getIDURL(),
        PacketID=stored_packet.PacketID,
        Payload=payload,
        RemoteID=recipient_idurl,
    )
    if _Debug:
        lg.args(_DebugLevel, file_size=sz, payload_size=len(payload), fn=filename, recipient=recipient_idurl)
    if recipient_idurl == stored_packet.OwnerID:
        if _Debug:
            lg.dbg(_DebugLevel, 'from request %r : sending %r back to owner: %s' % (newpacket, stored_packet, recipient_idurl))
        gateway.outbox(routed_packet)
        return True
    if _Debug:
        lg.dbg(_DebugLevel, 'from request %r : returning data owned by %s to %s' % (newpacket, stored_packet.OwnerID, recipient_idurl))
    gateway.outbox(routed_packet)
    return True


#------------------------------------------------------------------------------


def on_list_files(newpacket):
    json_query = {}
    try:
        j = serialization.BytesToDict(newpacket.Payload, keys_to_text=True, values_to_text=True)
        j['items'][0]
        json_query = j
    except:
        if strng.to_text(newpacket.Payload) == settings.ListFilesFormat():
            json_query = {
                'items': ['*'],
            }
    if json_query is None:
        lg.exc('unrecognized ListFiles() query received')
        return False
    # TODO: perform validations before sending back list of files
    list_files_global_id = global_id.ParseGlobalID(newpacket.PacketID)
    if list_files_global_id['key_id']:
        # customer id and data id can be recognized from packet id
        # return back list of files according to the request
        customer_idurl = list_files_global_id['idurl']
        key_id = list_files_global_id['key_id']
    else:
        # packet id format is unknown
        # by default returning back all files from that recipient if he is a customer
        customer_idurl = newpacket.OwnerID
        key_id = my_keys.make_key_id(alias='customer', creator_idurl=customer_idurl)
    key_id = my_keys.latest_key_id(key_id)
    list_files.send(
        customer_idurl=customer_idurl,
        packet_id=newpacket.PacketID,
        format_type=settings.ListFilesFormat(),
        key_id=key_id,
        remote_idurl=newpacket.OwnerID,  # send back to the requesting node
        query_items=json_query['items'],
    )
    if _Debug:
        lg.args(_DebugLevel, r=newpacket.OwnerID, c=customer_idurl, k=key_id, pid=newpacket.PacketID)
    return True


#------------------------------------------------------------------------------


def on_delete_file(newpacket):
    # TODO: call verify_ownership()
    # SECURITY
    if not newpacket.Payload:
        ids = [
            newpacket.PacketID,
        ]
    else:
        ids = strng.to_text(newpacket.Payload).split('\n')
    filescount = 0
    dirscount = 0
    lg.warn('going to erase files: %s' % ids)
    customer_id = global_id.UrlToGlobalID(newpacket.OwnerID)
    for pcktID in ids:
        glob_path = global_id.ParseGlobalID(pcktID)
        if not glob_path['customer']:
            glob_path = global_id.ParseGlobalID(customer_id + ':' + pcktID)
        if not glob_path['path']:
            lg.err('got incorrect PacketID')
            p2p_service.SendFail(newpacket, 'incorrect path')
            return False
        if customer_id != glob_path['customer']:
            # TODO: check that out if this actually suppose to be allowed
            lg.warn('trying to delete a file stored for another customer')
            continue
        # TODO: add validation of customerGlobID
        # TODO: process requests from another customer
        # SECURITY
        filename = make_valid_filename(newpacket.OwnerID, glob_path)
        if not filename:
            lg.warn('got empty filename, bad customer or wrong packetID?')
            p2p_service.SendFail(newpacket, 'not a customer, or file not found')
            return False
        if os.path.isfile(filename):
            try:
                os.remove(filename)
                filescount += 1
            except:
                lg.exc()
        elif os.path.isdir(filename):
            try:
                bpio._dir_remove(filename)
                dirscount += 1
            except:
                lg.exc()
        else:
            lg.warn('path was not found %s' % filename)
        do_notify_supplier_file_modified(glob_path['key_alias'], glob_path['path'], 'delete', newpacket.OwnerID, newpacket.CreatorID)
    p2p_service.SendAck(newpacket)
    if _Debug:
        lg.dbg(_DebugLevel, 'from [%s] with %d IDs, %d files and %d folders were removed' % (newpacket.OwnerID, len(ids), filescount, dirscount))
    return True


def on_delete_backup(newpacket):
    # TODO: call verify_ownership()
    # SECURITY
    if not newpacket.Payload:
        ids = [
            newpacket.PacketID,
        ]
    else:
        ids = strng.to_text(newpacket.Payload).split('\n')
    count = 0
    if _Debug:
        lg.args(_DebugLevel, ids=ids)
    customer_id = global_id.UrlToGlobalID(newpacket.OwnerID)
    for bkpID in ids:
        glob_path = global_id.ParseGlobalID(bkpID)
        if not glob_path['customer']:
            glob_path = global_id.ParseGlobalID(customer_id + ':' + bkpID)
        if not glob_path['path']:
            lg.err('got incorrect BackupID')
            p2p_service.SendFail(newpacket, 'incorrect backupID')
            return False
        if customer_id != glob_path['customer']:
            lg.warn('trying to delete file stored for another cusomer')
            continue
        # TODO: add validation of customerGlobID
        # TODO: process requests from another customer
        filename = make_valid_filename(newpacket.OwnerID, glob_path)
        if not filename:
            lg.warn('got empty filename, bad customer or wrong packetID?')
            p2p_service.SendFail(newpacket, 'not a customer, or file not found')
            return False
        if os.path.isdir(filename):
            try:
                bpio._dir_remove(filename)
                count += 1
            except:
                lg.exc()
        elif os.path.isfile(filename):
            try:
                os.remove(filename)
                count += 1
            except:
                lg.exc()
        else:
            lg.warn('path not found %s' % filename)
        do_notify_supplier_file_modified(glob_path['key_alias'], glob_path['path'], 'delete', newpacket.OwnerID, newpacket.CreatorID)
    p2p_service.SendAck(newpacket)
    if _Debug:
        lg.dbg(_DebugLevel, 'from [%s] with %d IDs, %d were removed' % (newpacket.OwnerID, len(ids), count))
    return True


#------------------------------------------------------------------------------


def on_customer_accepted(evt):
    customer_idurl = id_url.field(evt.data.get('idurl'))
    if not customer_idurl:
        lg.warn('unknown customer idurl in event data payload')
        return False
    customer_glob_id = global_id.idurl2glob(customer_idurl)
    queue_id = global_id.MakeGlobalQueueID(
        queue_alias='supplier-file-modified',
        owner_id=customer_glob_id,
        supplier_id=my_id.getGlobalID(),
    )
    if not p2p_queue.is_queue_exist(queue_id):
        # customer_key_id = global_id.MakeGlobalID(customer=customer_glob_id, key_alias='customer')
        # if my_keys.is_key_registered(customer_key_id):
        # TODO: re-think again about the customer key, do we really need it?
        if True:
            try:
                p2p_queue.open_queue(queue_id)
            except Exception as exc:
                lg.warn('failed to open queue %s : %s' % (queue_id, str(exc)))
        # else:
        #     lg.warn('customer key %r for supplier queue not registered' % customer_key_id)
    if p2p_queue.is_queue_exist(queue_id):
        if not p2p_queue.is_producer_exist(my_id.getGlobalID()):
            try:
                p2p_queue.add_producer(my_id.getGlobalID())
            except Exception as exc:
                lg.warn('failed to add producer: %s' % str(exc))
        if p2p_queue.is_producer_exist(my_id.getGlobalID()):
            if not p2p_queue.is_producer_connected(my_id.getGlobalID(), queue_id):
                try:
                    p2p_queue.connect_producer(my_id.getGlobalID(), queue_id)
                except Exception as exc:
                    lg.warn('failed to connect producer: %s' % str(exc))
            if p2p_queue.is_producer_connected(my_id.getGlobalID(), queue_id):
                if not p2p_queue.is_event_publishing(my_id.getGlobalID(), 'supplier-file-modified'):
                    try:
                        p2p_queue.start_event_publisher(my_id.getGlobalID(), 'supplier-file-modified')
                    except Exception as exc:
                        lg.warn('failed to start event publisher: %s' % str(exc))
    if _Debug:
        lg.args(_DebugLevel, c=customer_glob_id, q=queue_id)
    return True


def on_customer_terminated(evt):
    customer_idurl = evt.data.get('idurl')
    if not customer_idurl:
        lg.warn('unknown customer idurl in event data payload')
        return False
    customer_glob_id = global_id.idurl2glob(customer_idurl)
    queue_id = global_id.MakeGlobalQueueID(
        queue_alias='supplier-file-modified',
        owner_id=customer_glob_id,
        supplier_id=my_id.getGlobalID(),
    )
    # TODO: need to decide when to stop producing
    # might be that other customers needs that info still
    if p2p_queue.is_event_publishing(my_id.getGlobalID(), 'supplier-file-modified'):
        try:
            p2p_queue.stop_event_publisher(my_id.getGlobalID(), 'supplier-file-modified')
        except Exception as exc:
            lg.warn('failed to stop event publisher: %s' % str(exc))
    if p2p_queue.is_producer_connected(my_id.getGlobalID(), queue_id):
        try:
            p2p_queue.disconnect_producer(my_id.getGlobalID(), queue_id)
        except Exception as exc:
            lg.warn('failed to disconnect producer: %s' % str(exc))
    if p2p_queue.is_producer_exist(my_id.getGlobalID()):
        try:
            p2p_queue.remove_producer(my_id.getGlobalID())
        except Exception as exc:
            lg.warn('failed to remove producer: %s' % str(exc))
    if p2p_queue.is_queue_exist(queue_id):
        try:
            p2p_queue.close_queue(queue_id)
        except Exception as exc:
            lg.warn('failed to stop queue %s : %s' % (queue_id, str(exc)))
    if _Debug:
        lg.args(_DebugLevel, c=customer_glob_id, q=queue_id)
    return True


#------------------------------------------------------------------------------


def on_identity_url_changed(evt):
    old_idurl = id_url.field(evt.data['old_idurl'])
    # update customer idurl in "space" file
    contacts_changed = False
    for customer_idurl in contactsdb.customers():
        if old_idurl == customer_idurl:
            customer_idurl.refresh()
            contacts_changed = True
            lg.info('found customer idurl rotated : %r -> %r' % (evt.data['old_idurl'], evt.data['new_idurl']))
    if contacts_changed:
        contactsdb.save_customers()
    # update meta info for that customer
    meta_info_changed = False
    all_meta_info = contactsdb.read_customers_meta_info_all()
    for customer_idurl_bin in list(all_meta_info.keys()):
        if id_url.is_cached(old_idurl) and id_url.is_cached(customer_idurl_bin):
            if old_idurl == id_url.field(customer_idurl_bin):
                latest_customer_idurl_bin = id_url.field(customer_idurl_bin).to_bin()
                if latest_customer_idurl_bin != customer_idurl_bin:
                    all_meta_info[latest_customer_idurl_bin] = all_meta_info.pop(customer_idurl_bin)
                    meta_info_changed = True
                    lg.info('found customer idurl rotated in customers meta info : %r -> %r' % (latest_customer_idurl_bin, customer_idurl_bin))
    if meta_info_changed:
        contactsdb.write_customers_meta_info_all(all_meta_info)
    # update customer idurl in "space" file
    space_dict, free_space = accounting.read_customers_quotas()
    space_changed = False
    for customer_idurl_bin in list(space_dict.keys()):
        if id_url.is_cached(old_idurl) and id_url.is_cached(customer_idurl_bin):
            if id_url.field(customer_idurl_bin) == old_idurl:
                latest_customer_idurl_bin = id_url.field(customer_idurl_bin).to_bin()
                if latest_customer_idurl_bin != customer_idurl_bin:
                    space_dict[latest_customer_idurl_bin] = space_dict.pop(customer_idurl_bin)
                    space_changed = True
                    lg.info('found customer idurl rotated in customer quotas dictionary : %r -> %r' % (latest_customer_idurl_bin, customer_idurl_bin))
    if space_changed:
        accounting.write_customers_quotas(space_dict, free_space)
    # rename customer folder where I store all his files
    old_customer_dirname = str(global_id.UrlToGlobalID(evt.data['old_idurl']))
    new_customer_dirname = str(global_id.UrlToGlobalID(evt.data['new_idurl']))
    customers_dir = settings.getCustomersFilesDir()
    old_owner_dir = os.path.join(customers_dir, old_customer_dirname)
    new_owner_dir = os.path.join(customers_dir, new_customer_dirname)
    if os.path.isdir(old_owner_dir):
        try:
            bpio.move_dir_recursive(old_owner_dir, new_owner_dir)
            lg.info('copied %r into %r' % (old_owner_dir, new_owner_dir))
            if os.path.exists(old_owner_dir):
                bpio._dir_remove(old_owner_dir)
                lg.warn('removed %r' % old_owner_dir)
        except:
            lg.exc()
    # update customer idurl in "spaceused" file
    local_tester.TestSpaceTime()
    # TODO: reconnect "supplier-file-modified" consumers & producers
    return True


#------------------------------------------------------------------------------


def on_service_supplier_request(json_payload, newpacket, info):
    customer_idurl = newpacket.OwnerID
    customer_id = global_id.UrlToGlobalID(customer_idurl)
    bytes_for_customer = 0
    try:
        bytes_for_customer = int(json_payload['needed_bytes'])
    except:
        lg.exc()
        return p2p_service.SendFail(newpacket, 'invalid payload')
    try:
        customer_public_key = json_payload['customer_public_key']
        customer_public_key_id = customer_public_key['key_id']
    except:
        customer_public_key = None
        customer_public_key_id = None
    data_owner_idurl = None
    target_customer_idurl = None
    family_position = json_payload.get('position')
    ecc_map = json_payload.get('ecc_map')
    family_snapshot = json_payload.get('family_snapshot')
    if family_snapshot:
        family_snapshot = id_url.to_bin_list(family_snapshot)
    target_customer_id = json_payload.get('customer_id')
    key_id = my_keys.latest_key_id(json_payload.get('key_id'))
    if key_id:
        # this is a request from external user to access shared data stored by one of my customers
        # this is "second" customer requesting data from "first" customer
        if not key_id or not my_keys.is_valid_key_id(key_id):
            lg.warn('missed or invalid key id')
            return p2p_service.SendFail(newpacket, 'invalid key id')
        target_customer_idurl = global_id.GlobalUserToIDURL(target_customer_id)
        if not contactsdb.is_customer(target_customer_idurl):
            lg.warn('target user %s is not a customer' % target_customer_id)
            return p2p_service.SendFail(newpacket, 'not a customer')
        if target_customer_idurl == customer_idurl:
            lg.warn('customer %s requesting shared access to own files' % customer_idurl)
            return p2p_service.SendFail(newpacket, 'invalid case')
        if not my_keys.is_key_registered(key_id):
            lg.warn('key not registered: %s' % key_id)
            p2p_service.SendFail(newpacket, 'key not registered')
            return False
        data_owner_idurl = my_keys.split_key_id(key_id)[1]
        if not id_url.is_the_same(data_owner_idurl, target_customer_idurl) and not id_url.is_the_same(data_owner_idurl, customer_idurl):
            # pretty complex scenario:
            # external customer requesting access to data which belongs not to that customer
            # this is "third" customer accessing data belongs to "second" customer
            # TODO: for now just refuse it
            lg.warn('under construction, key_id=%s customer_idurl=%s target_customer_idurl=%s' % (key_id, customer_idurl, target_customer_idurl))
            p2p_service.SendFail(newpacket, 'under construction')
            return False
        register_customer_key(customer_public_key_id, customer_public_key)
        # do not create connection with that customer, only accept the request
        lg.info('external customer %s requested access to shared data at %s' % (customer_id, key_id))
        return p2p_service.SendAck(newpacket, 'accepted')
    # key_id is not present in the request:
    # this is a request to connect new customer (or reconnect existing one) to that supplier
    if not bytes_for_customer or bytes_for_customer < 0:
        lg.warn('wrong payload : %s' % newpacket.Payload)
        return p2p_service.SendFail(newpacket, 'wrong storage value')
    current_customers = contactsdb.customers()
    if accounting.check_create_customers_quotas():
        lg.info('created new customers quotas file')
    space_dict, free_space = accounting.read_customers_quotas()
    try:
        free_bytes = int(free_space)
    except:
        lg.exc()
        return p2p_service.SendFail(newpacket, 'broken quotas file')
    if (customer_idurl not in current_customers and customer_idurl.to_bin() in list(space_dict.keys())):
        lg.err('broken quotas file: %r is not a customer, but found in the quotas file' % customer_idurl.to_bin())
        return p2p_service.SendFail(newpacket, 'broken quotas file')
    if (customer_idurl in current_customers and customer_idurl.to_bin() not in list(space_dict.keys())):
        # seems like customer's idurl was rotated, but space file still have the old idurl
        # need to find that old idurl value and replace with the new one
        for other_customer_idurl in space_dict.keys():
            if other_customer_idurl and other_customer_idurl != 'free' and id_url.field(other_customer_idurl) == customer_idurl:
                lg.info('found rotated customer identity in space file, switching: %r -> %r' % (other_customer_idurl, customer_idurl.to_bin()))
                space_dict[customer_idurl.to_bin()] = space_dict.pop(other_customer_idurl)
                break
        if customer_idurl.to_bin() not in list(space_dict.keys()):
            lg.err('broken customers file: %r is a customer, but not found in the quotas file' % customer_idurl.to_bin())
            return p2p_service.SendFail(newpacket, 'broken customers file')
    # check/verify/create/update contract with requestor customer
    # the contracts are needed to keep track of consumed resources
    current_contract = {}
    if driver.is_on('service_supplier_contracts'):
        from bitdust.supplier import storage_contract
        try:
            current_contract = storage_contract.prepare_customer_contract(customer_idurl, details={
                'allocated_bytes': bytes_for_customer,
                'ecc_position': family_position,
                'ecc_map': ecc_map,
            })
        except:
            lg.exc()
            current_contract = {}
        if current_contract and current_contract.get('deny'):
            lg.warn('contract processing denied with user %s' % customer_idurl)
            # TODO: disabled for now...
            current_contract = {}
            # return p2p_service.SendFail(newpacket, 'deny:' + jsn.dumps(current_contract))
    # check if this is a new customer or an existing one
    # for existing one, we have to first release currently allocated resources
    if customer_idurl in current_customers:
        current_allocated_byes = int(space_dict.get(customer_idurl.to_bin(), 0))
        free_bytes += current_allocated_byes
        current_customers.remove(customer_idurl)
        space_dict.pop(customer_idurl.to_bin())
        new_customer = False
    else:
        new_customer = True
    # lg.args(8, new_customer=new_customer, current_allocated_bytes=space_dict.get(customer_idurl.to_bin()))
    if free_bytes <= bytes_for_customer:
        contactsdb.remove_customer_meta_info(customer_idurl)
        accounting.write_customers_quotas(space_dict, free_bytes)
        contactsdb.update_customers(current_customers)
        contactsdb.save_customers()
        if customer_public_key_id:
            my_keys.erase_key(customer_public_key_id)
        reactor.callLater(0, local_tester.TestUpdateCustomers)  # @UndefinedVariable
        if new_customer:
            lg.info('NEW CUSTOMER: DENIED     not enough space available')
            events.send('new-customer-denied', data=dict(idurl=customer_idurl))
        else:
            lg.info('OLD CUSTOMER: DENIED     not enough space available')
            events.send('existing-customer-denied', data=dict(idurl=customer_idurl))
        return p2p_service.SendAck(newpacket, 'deny:{"reason": "not enough space available"}')
    free_bytes = free_bytes - bytes_for_customer
    current_customers.append(customer_idurl)
    space_dict[customer_idurl.to_bin()] = bytes_for_customer
    contactsdb.add_customer_meta_info(customer_idurl, {
        'ecc_map': ecc_map,
        'position': family_position,
        'family_snapshot': family_snapshot,
    })
    accounting.write_customers_quotas(space_dict, free_bytes)
    contactsdb.update_customers(current_customers)
    contactsdb.save_customers()
    register_customer_key(customer_public_key_id, customer_public_key)
    reactor.callLater(0, local_tester.TestUpdateCustomers)  # @UndefinedVariable
    if new_customer:
        lg.info('NEW CUSTOMER: ACCEPTED   %s family_position=%s ecc_map=%s allocated_bytes=%s' % (customer_idurl, family_position, ecc_map, bytes_for_customer))
        events.send(
            'new-customer-accepted', data=dict(
                idurl=customer_idurl,
                allocated_bytes=bytes_for_customer,
                ecc_map=ecc_map,
                position=family_position,
                family_snapshot=family_snapshot,
                key_id=customer_public_key_id,
                contract=current_contract,
            )
        )
    else:
        lg.info('EXISTING CUSTOMER: ACCEPTED  %s family_position=%s ecc_map=%s allocated_bytes=%s' % (customer_idurl, family_position, ecc_map, bytes_for_customer))
        events.send(
            'existing-customer-accepted', data=dict(
                idurl=customer_idurl,
                allocated_bytes=bytes_for_customer,
                ecc_map=ecc_map,
                position=family_position,
                key_id=customer_public_key_id,
                family_snapshot=family_snapshot,
                contract=current_contract,
            )
        )
    if current_contract:
        return p2p_service.SendAck(newpacket, 'accepted:' + jsn.dumps(current_contract))
    return p2p_service.SendAck(newpacket, 'accepted')


def on_service_supplier_cancel(json_payload, newpacket, info):
    customer_idurl = newpacket.OwnerID
    try:
        customer_public_key = json_payload['customer_public_key']
        customer_public_key_id = customer_public_key['key_id']
    except:
        customer_public_key = None
        customer_public_key_id = None
    customer_ecc_map = json_payload.get('ecc_map')
    if driver.is_on('service_supplier_contracts'):
        from bitdust.supplier import storage_contract
        storage_contract.cancel_customer_contract(customer_idurl)
    if not contactsdb.is_customer(customer_idurl):
        lg.warn('got packet from %s, but he is not a customer' % customer_idurl)
        return p2p_service.SendFail(newpacket, 'not a customer')
    if accounting.check_create_customers_quotas():
        lg.info('created a new space file')
    space_dict, free_space = accounting.read_customers_quotas()
    if customer_idurl.to_bin() not in list(space_dict.keys()):
        lg.warn('got packet from %s, but not found him in space dictionary' % customer_idurl)
        return p2p_service.SendFail(newpacket, 'not a customer')
    try:
        free_bytes = int(free_space)
        free_space = free_bytes + int(space_dict[customer_idurl.to_bin()])
    except:
        lg.exc()
        return p2p_service.SendFail(newpacket, 'broken quotas file')
    new_customers = list(contactsdb.customers())
    new_customers.remove(customer_idurl)
    space_dict.pop(customer_idurl.to_bin())
    accounting.write_customers_quotas(space_dict, free_space)
    contactsdb.remove_customer_meta_info(customer_idurl)
    contactsdb.update_customers(new_customers)
    contactsdb.save_customers()
    if customer_public_key_id:
        my_keys.erase_key(customer_public_key_id)
    # TODO: erase customer's groups keys also
    reactor.callLater(0, local_tester.TestUpdateCustomers)  # @UndefinedVariable
    lg.info('EXISTING CUSTOMER TERMINATED %r' % customer_idurl)
    events.send('existing-customer-terminated', data=dict(idurl=customer_idurl, ecc_map=customer_ecc_map))
    return p2p_service.SendAck(newpacket, 'accepted')


#------------------------------------------------------------------------------


def on_inbox_packet_received(newpacket, info, status, error_message):
    if newpacket.Command == commands.DeleteFile():
        return on_delete_file(newpacket)
    elif newpacket.Command == commands.DeleteBackup():
        return on_delete_backup(newpacket)
    elif newpacket.Command == commands.Retrieve():
        return on_retrieve(newpacket)
    elif newpacket.Command == commands.Data():
        return on_data(newpacket)
    elif newpacket.Command == commands.ListFiles():
        return on_list_files(newpacket)
    return False


def on_suppliers_dht_layer_connected(ok):
    lg.info('connected to DHT layer for suppliers: %r' % ok)
    if my_id.getIDURL():
        dht_service.set_node_data('idurl', my_id.getIDURL().to_text(), layer_id=dht_records.LAYER_SUPPLIERS)
    return ok


def on_dht_layer_connected(evt):
    if evt.data['layer_id'] == 0:
        connect_suppliers_dht_layer()
    elif evt.data['layer_id'] == dht_records.LAYER_SUPPLIERS:
        on_suppliers_dht_layer_connected(True)
